#!/usr/bin/env python3
#
# Tests if the myokit.lib.plots module runs without exceptions, doesn't inspect
# the output.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit
import myokit.lib.plots as plots

from myokit.tests import DIR_DATA, WarningCollector


class LibPlotTest(unittest.TestCase):
    """
    Tests if the myokit.lib.plots module runs without exceptions, doesn't
    inspect the output.
    """

    def test_simulation_times(self):
        # Test the simulation times() plots (has several modes).

        # Select matplotlib backend that doesn't require a screen
        import matplotlib
        matplotlib.use('template')
        import matplotlib.pyplot as plt

        # Run simulation, get benchmarks
        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        s = myokit.Simulation(m, p)
        d = s.run(1000, log=[
            'engine.time', 'engine.realtime', 'engine.evaluations'])
        st = d['engine.time']
        rt = d['engine.realtime']
        ev = d['engine.evaluations']

        # Stair
        fig = plt.figure()
        plots.simulation_times(st, rt, ev, 'stair')
        plt.show()
        plt.close(fig)
        self.assertRaises(ValueError, plots.simulation_times, mode='stair')

        # Inverse stair
        fig = plt.figure()
        plots.simulation_times(st, rt, ev, 'stair_inverse')
        plt.show()
        plt.close(fig)
        self.assertRaises(
            ValueError, plots.simulation_times, mode='stair_inverse')

        # Load
        fig = plt.figure()
        plots.simulation_times(st, rt, ev, 'load')
        plt.show()
        plt.close(fig)
        self.assertRaises(ValueError, plots.simulation_times, mode='load')

        # Histogram
        fig = plt.figure()
        plots.simulation_times(st, rt, ev, 'histo')
        plt.show()
        plt.close(fig)
        self.assertRaises(ValueError, plots.simulation_times, mode='histo')

        # Time per step
        fig = plt.figure()
        plots.simulation_times(st, rt, ev, 'time_per_step')
        plt.show()
        plt.close(fig)
        self.assertRaises(
            ValueError, plots.simulation_times, mode='time_per_step')

        # Evaluations per step
        fig = plt.figure()
        plots.simulation_times(st, rt, ev, 'eval_per_step')
        plt.show()
        plt.close(fig)
        self.assertRaises(
            ValueError, plots.simulation_times, mode='eval_per_step')

        # Non-existing model
        self.assertRaises(
            ValueError, plots.simulation_times, st, rt, ev, mode='xxx')

    def test_current_arrows(self):
        # Test the current arrows plot.

        # Select matplotlib backend that doesn't require a screen
        import matplotlib
        matplotlib.use('template')
        import matplotlib.pyplot as plt

        # Run simulation, get currents
        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))

        # Add dummy current, that never carries much charge and should be
        # ignored
        c = m.add_component('dummy')
        x = c.add_variable('IDummy')
        x.set_rhs('0 * membrane.V')

        s = myokit.Simulation(m, p)
        currents = ['ina.INa', 'ik1.IK1', 'ica.ICa', 'dummy.IDummy']
        d = s.run(600, log=['engine.time', 'membrane.V'] + currents)

        fig = plt.figure()
        plots.current_arrows(d, 'membrane.V', currents)
        plt.close(fig)
        plt.show()

        # Massive peak at final point
        d = d.npview()
        d['dummy.IDummy'][-1] = 100
        fig = plt.figure()
        plots.current_arrows(d, 'membrane.V', currents)
        plt.close(fig)
        plt.show()

    def test_cumulative_current(self):
        # Test the cumulative current plot.

        # Select matplotlib backend that doesn't require a screen
        import matplotlib
        matplotlib.use('template')
        import matplotlib.pyplot as plt

        # Run simulation, get currents
        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        s = myokit.Simulation(m, p)
        currents = ['ica.ICa', 'ik.IK', 'ik1.IK1', 'ikp.IKp', 'ib.Ib']
        d = s.run(600, log=['engine.time', 'membrane.V'] + currents)

        # No extra arguments
        fig = plt.figure()
        plots.cumulative_current(d, currents)
        plt.legend()
        plt.close(fig)
        plt.show()

        # Labels set
        labels = ['I_Ca', 'I_K', 'I_K1', 'I_Kp', 'I_b']
        fig = plt.figure()
        plots.cumulative_current(d, currents, labels=labels)
        plt.legend()
        plt.close(fig)
        plt.show()

        # Colors set
        colors = ['green', 'blue', 'yellow', 'brown', 'gray']
        fig = plt.figure()
        plots.cumulative_current(d, currents, colors=colors)
        plt.legend()
        plt.close(fig)
        plt.show()

        # Not enough colors set (will repeat array)
        colors = ['green', 'blue']
        fig = plt.figure()
        plots.cumulative_current(d, currents, colors=colors)
        plt.legend()
        plt.close(fig)
        plt.show()

        # Integrate currents to charges
        fig = plt.figure()
        plots.cumulative_current(d, currents, integrate=True)
        plt.legend()
        plt.close(fig)
        plt.show()

        # Normalize currents
        fig = plt.figure()
        plots.cumulative_current(d, currents, normalize=True)
        plt.legend()
        plt.close(fig)
        plt.show()

        # Normalize currents and set maximum number of currents shown
        fig = plt.figure()
        plots.cumulative_current(d, currents, normalize=True, max_currents=3)
        plt.legend()
        plt.close(fig)
        plt.show()

        with WarningCollector() as w:
            fig = plt.figure()
            plots.cumulative_current(d, currents, normalise=True)
            plt.legend()
            plt.close(fig)
            plt.show()
        self.assertIn('deprecated', w.text())


if __name__ == '__main__':
    unittest.main()



#
# Provides HTML support.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._exporter import HTMLExporter
from ._flatten import html2ascii    # noqa


# Importers
# Exporters
_exporters = {
    'html': HTMLExporter,
}


def exporters():
    """
    Returns a dict of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
# Language keywords
# None!



#
# SetupTools script for Myokit
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from setuptools import setup, find_packages


# Get version number
import os
import sys
sys.path.append(os.path.abspath('myokit'))
from _myokit_version import __version__ as version  # noqa
sys.path.pop()
del os, sys


# Load text for description and license
with open('README.md') as f:
    readme = f.read()


# Go!
setup(
    # See https://python-packaging.readthedocs.io/en/latest/index.html
    # for details of what goes in here.

    # Module name (lowercase)
    name='myokit',

    # Version
    version=version,

    # Description
    description='A simple interface to cardiac cellular electrophysiology',
    long_description=readme,
    long_description_content_type='text/markdown',

    # Author and license
    license='BSD 3-clause license',
    author='Michael Clerx',
    author_email='michael@myokit.org',

    # URLs
    url='http://myokit.org',
    project_urls={
        'Bug Tracker': 'https://github.com/MichaelClerx/myokit/issues',
        'Documentation': 'http://docs.myokit.org',
        'Source Code': 'https://github.com/MichaelClerx/myokit',
    },

    # Classifiers for pypi
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Physics',
    ],

    # Packages to include
    packages=find_packages(include=('myokit', 'myokit.*')),

    # Include non-python files (via MANIFEST.in)
    include_package_data=True,

    # Register myokit as a shell script
    entry_points={
        'console_scripts': ['myokit = myokit.__main__:main']
    },

    # Python version
    python_requires='>=3.7',

    # List of dependencies
    install_requires=[
        'configparser',
        'lxml',
        'matplotlib>=2.2',
        'numpy',
        'setuptools',
        # PyQT or PySide?
        # (PySide is pip installable, Actions can get PyQt from apt)
    ],

    # Optional extras
    extras_require={
        'docs': [
            'sphinx>=1.7.4',        # Doc generation
        ],
        'dev': [
            'coverage',             # Coverage checking
            'flake8>=3',            # Style checking
        ],
        'optional': [
            'scipy',            # Used in data log, and data log viewer
            'sympy',            # Used in formats.sympy
        ],
        'gui': ['pyqt6', 'sip'],
        'pyqt': ['pyqt6', 'sip'],
        'pyside': ['pyside6'],
    },

    # Unit tests
    test_suite='myokit.tests',

    # See: https://setuptools.pypa.io/en/latest/userguide/miscellaneous.html
    #      #setting-the-zip-safe-flag
    zip_safe=False,
)



#
# OpenCL driven fiber-tissue simulation.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import platform

import myokit

# Location of C and OpenCL sources
SOURCE_FILE = 'fiber_tissue.c'
KERNEL_FILE = 'openclsim.cl'


class FiberTissueSimulation(myokit.CModule):
    """
    Runs a simulation of a fiber leading up to a rectangular piece of tissue.

    Takes the following input arguments:

    ``fiber_model``
        The model to simulate the fiber with.
    ``tissue_model``
        The model to simulate the tissue with. Both models will be cloned when
        the simulation is created so that no changes to the given models will
        be made.
    ``protocol``
        An optional pacing protocol, used to stimulate a number of cells at the
        start of the fiber.
    ``ncells_fiber``
        The number of cells in the fiber (a tuple).
    ``ncells_tissue``
        The number of cells in the tissue (a tuple).
    ``nx_paced``
        The width (in cells) of the stimulus applied to the fiber. The fiber
        will be stimulated along its full height.
    ``g_fiber``
        The cell to cell conductance in the fiber (a tuple).
    ``g_tissue``
        The cell to cell conductance in the tissue (a tuple).
    ``g_fiber_tissue``
        The fiber-cell to tissue-cell conductance at the junction (a scalar).
    ``dt``
        The time step to use in the forward-Euler integration scheme.
    ``precision``
        Can be set to ``myokit.SINGLE_PRECISION`` (default) or
        ``myokit.DOUBLE_PRECISION`` if the used device supports it.
    ``native_maths``
        On some devices, selected functions (e.g. ``exp``) can be made to run
        faster (but possibly less accurately) by setting ``native_maths=True``.

    The simulation provides the following inputs variables can bind to:

    ``time`` (global)
        The simulation time
    ``pace`` (per-cell)
        The pacing level, this is set if a protocol was passed in. Will be set
        to 0 if no protocol is provided.
    ``diffusion_current`` (per-cell)
        The current flowing from the cell to its neighbors. This will be
        positive when the cell is acting as a source, negative when it is
        acting as a sink. Will be set to 0 if no connections are made.

    The variable ``time`` is set globally, meaning each cell uses the same
    value. The variables ``pace`` and ``diffusion_current`` have different
    values per cell.

    The following labeled variables are required for this simulation to work:

    ``membrane_potential``
        The variable representing the membrane potential.

    The ``diffusion_current`` is calculated as::

        i = gx * ((V - V_xnext) - (V_xlast - V))
          + gy * ((V - V_ynext) - (V_ylast - V))

    At the boundaries, where either ``V_ilast`` or ``V_inext`` is unavailable,
    the value of ``V`` is substituted, causing the term to go to zero. The
    values of ``gx`` and ``gy`` can be set in the simulation's constructor.

    For a typical model with currents in ``[uA/uF]`` and voltage in ``[mV]``,
    `gx`` and ``gy`` have the unit ``[mS/uF]``.

    Simulations maintain an internal state consisting of

    - the current simulation time
    - the current states
    - the default states

    When a simulation is created, the simulation time is set to 0 and both the
    current and the default state are equal to the state of the given models,
    copied once for each cell.
    After each call to :meth:`run` the time variable and current state are
    updated, so that each successive call to run continues where the previous
    simulation left off. A :meth:`reset` method is provided that will set the
    time back to 0 and revert the current state to the default state. To change
    the time or state manually, use :meth:`set_time` and
    :meth:`set_fiber_state` and :meth:`set_tissue_state`.

    A pre-pacing method :meth:`pre` is provided that doesn't affect the
    simulation time but will update the current *and the default state*. This
    allows you to pre-pace, run a simulation, reset to the pre-paced state, run
    another simulation etc.

    The model passed to the simulation is cloned and stored internally, so
    changes to the original model object will not affect the simulation.

    Models used with this simulation need to have independent components: it
    should be possible to evaluate the model's equations one component at a
    time. A model's suitability can be tested using :meth:`
    has_interdependent_components
    <myokit.Model.has_interdependent_components>`.
    """
    _index = 0  # Unique id for the created simulation module

    def __init__(
            self, fiber_model, tissue_model, protocol=None,
            ncells_fiber=(128, 2), ncells_tissue=(128, 128), nx_paced=5,
            g_fiber=(9, 6), g_tissue=(9, 6), g_fiber_tissue=9,
            dt=0.005, precision=myokit.SINGLE_PRECISION, native_maths=False):
        super().__init__()

        # Deprecated since 2021-05-28
        # Un-deprecated on 2022-04-20, following Adam's request
        #import warnings
        #warnings.warn(
        #    'The class `myokit.FiberTissueSimulation` is deprecated, and will'
        #    ' be removed in future versions of Myokit.')

        # List of globally logged inputs
        self._global = ['time', 'pace']

        # Require valid models
        fiber_model.validate()
        tissue_model.validate()

        # Require independent components
        if fiber_model.has_interdependent_components():
            cycles = '\n'.join([
                '  ' + ' > '.join([x.name() for x in c])
                for c in fiber_model.component_cycles()])
            raise ValueError(
                'This simulation requires models without interdependent'
                ' components. Please restructure the fiber model and re-run.'
                ' Cycles:\n' + cycles)
        if tissue_model.has_interdependent_components():
            cycles = '\n'.join([
                '  ' + ' > '.join([x.name() for x in c])
                for c in tissue_model.component_cycles()])
            raise ValueError(
                'This simulation requires models without interdependent'
                ' components. Please restructure the tissue model and re-run.'
                ' Cycles:\n' + cycles)

        # Clone models, store
        fiber_model = fiber_model.clone()
        tissue_model = tissue_model.clone()
        self._modelf = fiber_model
        self._modelt = tissue_model

        # Set protocol
        self.set_protocol(protocol)

        # Check dimensionality, number of cells
        msg = 'The fiber size must be a tuple (nx, ny).'
        try:
            if len(ncells_fiber) != 2:
                raise ValueError(msg)
        except TypeError:
            raise ValueError(msg)
        msg = 'The tissue size must be a tuple (nx, ny).'
        try:
            if len(ncells_tissue) != 2:
                raise ValueError(msg)
        except TypeError:
            raise ValueError(msg)
        self._ncellsf = [int(x) for x in ncells_fiber]
        self._ncellst = [int(x) for x in ncells_tissue]
        if self._ncellsf[0] < 1 or self._ncellsf[1] < 1:
            raise ValueError('The fiber size must be at least (1, 1).')
        if self._ncellst[0] < 1 or self._ncellst[1] < 1:
            raise ValueError('The tissue size must be at least (1, 1).')
        if self._ncellsf[1] > self._ncellst[1]:
            raise ValueError(
                'The fiber y-dimension cannot exceed that of the tissue.')

        # Check width of pacing stimulus
        nx_paced = int(nx_paced)
        if nx_paced < 0:
            raise ValueError(
                'The width of the stimulus pulse must be non-negative.')
        nx_paced = min(nx_paced, self._ncellsf[0])
        self._paced_cells = []
        for y in range(self._ncellsf[1]):
            for x in range(min(self._ncellsf[0], nx_paced)):
                self._paced_cells.append(x + y * self._ncellsf[0])

        # Check conductivities
        msg = 'The fiber conductivity must be a tuple (gx, gy).'
        try:
            if len(g_fiber) != 2:
                raise ValueError(msg)
        except TypeError:
            raise ValueError(msg)
        msg = 'The tissue conductivity must be a tuple (gx, gy).'
        try:
            if len(g_tissue) != 2:
                raise ValueError(msg)
        except TypeError:
            raise ValueError(msg)
        self._gf = [float(x) for x in g_fiber]
        self._gt = [float(x) for x in g_tissue]
        self._gft = float(g_fiber_tissue)

        # Point of connection to tissue
        self._cfx = self._ncellsf[0] - 1
        self._ctx = 0
        self._cty = int(0.5 * (self._ncellst[1] - self._ncellsf[1]))

        # Check step size
        dt = float(dt)
        if dt <= 0:
            raise ValueError('The step size must be greater than zero.')
        self._step_size = dt

        # Set precision
        if precision not in (myokit.SINGLE_PRECISION, myokit.DOUBLE_PRECISION):
            raise ValueError('Only single and double precision are supported.')
        self._precision = precision

        # Set native maths
        self._native_math = bool(native_maths)

        # Set remaining properties
        self._time = 0
        self._nstatef = self._modelf.count_states()
        self._nstatet = self._modelt.count_states()
        self._ntotalf = self._ncellsf[0] * self._ncellsf[1]
        self._ntotalt = self._ncellst[0] * self._ncellst[1]

        # Get membrane potential variables
        self._vmf = self._modelf.label('membrane_potential')
        if self._vmf is None:
            raise ValueError(
                'This simulation requires the membrane potential variable to'
                ' be labelled as "membrane_potential" in the fiber model.')
        if not self._vmf.is_state():
            raise ValueError(
                'The variable labelled as membrane potential in the fiber'
                ' model must be a state variable.')
        self._vmt = self._modelt.label('membrane_potential')
        if self._vmt is None:
            raise ValueError(
                'This simulation requires the membrane potential variable to'
                ' be labelled as "membrane_potential" in the tissue model.')
        if not self._vmt.is_state():
            raise ValueError(
                'The variable labelled as membrane potential in the tissue'
                ' model must be a state variable.')

        # Check for binding to diffusion_current
        if self._modelf.binding('diffusion_current') is None:
            raise ValueError(
                'This simulation requires a variable in the fiber model to be'
                ' bound to "diffusion_current" to pass current from one cell'
                ' to the next.')
        if self._modelt.binding('diffusion_current') is None:
            raise ValueError(
                'This simulation requires a variable in the tissue model to be'
                ' bound to "diffusion_current" to pass current from one cell'
                ' to the next.')

        # Check if membrane potentials have same unit
        uvf = self._vmf.unit()
        if uvf is None:
            raise ValueError('The fiber model must specify a unit for the'
                             ' membrane potential.')
        uvt = self._vmt.unit()
        if uvt is None:
            raise ValueError('The tissue model must specify a unit for the'
                             ' membrane potential.')
        if uvf != uvt:
            raise ValueError(
                'The membrane potential must have the same unit in the fiber'
                ' and the tissue model: ' + str(uvf) + ' vs ' + str(uvt) + '.')

        # Check if diffusion current variables have same unit
        ucf = self._modelf.binding('diffusion_current').unit()
        uct = self._modelt.binding('diffusion_current').unit()
        if ucf is None:
            raise ValueError('The fiber model must specify a unit for the'
                             ' diffusion current.')
        if uct is None:
            raise ValueError('The tissue model must specify a unit for the'
                             ' diffusion current.')
        if ucf != uct:
            raise ValueError(
                'The diffusion current must have the same unit in the fiber'
                ' and the tissue model: ' + str(ucf) + ' vs ' + str(uct) + '.')

        # Set state and default state
        self._statef = self._modelf.initial_values(True) * self._ntotalf
        self._statet = self._modelt.initial_values(True) * self._ntotalt
        self._default_statef = list(self._statef)
        self._default_statet = list(self._statet)

        # Process bindings, remove unsupported bindings, get map of bound
        # variables to internal names.
        self._bound_variablesf = myokit._prepare_bindings(self._modelf, {
            'time': 'time',
            'pace': 'pace',
            'diffusion_current': 'idiff',
        })
        self._bound_variablest = myokit._prepare_bindings(self._modelt, {
            'time': 'time',
            'pace': 'pace',
            'diffusion_current': 'idiff',
        })

        # Create unique names
        self._modelf.create_unique_names()
        self._modelt.create_unique_names()

        # Create back-end
        self._create_backend()

    def _create_backend(self):
        """
        Creates this simulation's backend.
        """
        # Unique simulation id
        FiberTissueSimulation._index += 1
        mname = 'myokit_sim_fiber_tissue_' + str(FiberTissueSimulation._index)
        mname += '_' + str(myokit.pid_hash())

        # Arguments
        args = {
            'module_name': mname,
            'modelf': self._modelf,
            'modelt': self._modelt,
            'vmf': self._vmf,
            'vmt': self._vmt,
            'boundf': self._bound_variablesf,
            'boundt': self._bound_variablest,
            'precision': self._precision,
            'native_math': self._native_math,
        }
        fname = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)

        # Define libraries
        libs = []
        flags = []
        plat = platform.system()
        if plat != 'Darwin':    # pragma: no macos cover
            libs.append('OpenCL')
        else:                   # pragma: no cover
            flags.append('-framework')
            flags.append('OpenCL')
        if plat != 'Windows':   # pragma: no windows cover
            libs.append('m')

        # Create extension
        libd = list(myokit.OPENCL_LIB)
        incd = list(myokit.OPENCL_INC)
        incd.append(myokit.DIR_CFUNC)
        self._sim = self._compile(
            mname, fname, args, libs, libd, incd, larg=flags,
            continue_in_debug_mode=True)

    def default_fiber_state(self, x=None, y=None):
        """
        Returns the current default simulation state for the fiber model as a
        list of ``len(state) * n_total_cells`` floating point values, where
        ``n_total_cells`` is the total number of fiber cells.

        If the optional arguments ``x`` and ``y`` specify a valid cell index, a
        single cell's state is returned.

        The returned list is indexed so that x changes first (the first ``nx``
        entries have ``y = 0``, the second ``nx`` entries have ``y = 1`` and so
        on).
        """
        if x is None:
            return list(self._default_statef)
        elif y is None:
            raise ValueError('default_fiber_state() must be called with no'
                             ' arguments or with both an x and y argument.')
        else:
            nx, ny = self._ncellsf
            x = int(x or 0)
            if x < 0 or x >= nx:
                raise IndexError('Given x-index out of range.')
            y = int(y or 0)
            if y < 0 or y >= ny:
                raise IndexError('Given y-index out of range.')
            x += y * nx
            return self._default_statef[
                x * self._nstatef:(x + 1) * self._nstatef]

    def default_tissue_state(self, x=None, y=None):
        """
        Returns the current default simulation state for the tissue model as a
        list of ``len(state) * n_total_cells`` floating point values, where
        ``n_total_cells`` is the total number of tissue cells.

        If the optional arguments ``x`` and ``y`` specify a valid cell index, a
        single cell's state is returned.

        The returned list is indexed so that x changes first (the first ``nx``
        entries have ``y = 0``, the second ``nx`` entries have ``y = 1`` and so
        on).
        """
        if x is None:
            return list(self._default_statet)
        elif y is None:
            raise ValueError('default_tissue_state() must be called with no'
                             ' arguments or with both an x and y argument.')
        else:
            nx, ny = self._ncellst
            x = int(x or 0)
            if x < 0 or x >= nx:
                raise IndexError('Given x-index out of range.')
            y = int(y or 0)
            if y < 0 or y >= ny:
                raise IndexError('Given y-index out of range.')
            x += y * nx
            return self._default_statet[
                x * self._nstatet:(x + 1) * self._nstatet]

    def fiber_shape(self):
        """
        Returns the shape of this simulation's grid of fiber cells as a tuple
        ``(ny, nx)``.
        """
        # TODO: See https://github.com/MichaelClerx/myokit/issues/764
        return (self._ncellsf[1], self._ncellsf[0])

    def fiber_state(self, x=None, y=None):
        """
        Returns the current simulation state in the fiber as a list of
        ``len(state_fiber) * ncells_fiber`` floating point values.

        If the optional arguments ``x`` and ``y`` specify a valid cell index a
        single fiber cell's state is returned.
        """
        if x is None:
            return list(self._statef)
        elif y is None:
            raise ValueError('fiber_state() must be called with no arguments'
                             ' or with both an x and y argument.')
        else:
            x = int(x)
            if x < 0 or x >= self._ncellsf[0]:
                raise IndexError('Given x-index out of range.')
            y = int(y)
            if y < 0 or y >= self._ncellsf[1]:
                raise IndexError('Given y-index out of range.')
            x += y * self._ncellsf[0]
            return self._statef[x * self._nstatef:(x + 1) * self._nstatef]

    def find_nan(self, logf, logt):
        """
        Searches for the origin of a ``NaN`` (or ``inf``) in a set of
        simulation logs generated by this Simulation.

        The logs must contain the state of each cell and all bound variables.
        The NaN can occur at any point in time except the first.

        Returns a tuple ``(part, time, icell, variable, value, states, bound)``
        where ``time`` is the time the first ``NaN`` was found and ``icell`` is
        the index of the cell in which it happened. The entry ``part`` is a
        string containing either "fiber" or "tissue", indicating which part of
        the simulation triggered the error. The offending variable's name is
        given as ``variable`` and its (illegal) value as ``value``. The current
        state and, if available, any previous states are given in the list
        ``states``. Here, ``states[0]`` points to the current state in the
        simulation part causing the error, ``state[1]`` is the previous state
        and so on. Similarly the values of the error causing model's bound
        variables is given in ``bound``.
        """
        import numpy as np
        # Check if logs contain all states and bound variables
        lt = []
        lf = []
        for label in self._global:
            v = self._modelf.binding(label)
            if v is not None:
                lf.append(v.qname())
            v = self._modelt.binding(label)
            if v is not None:
                lt.append(v.qname())
        lf = myokit.prepare_log(
            myokit.LOG_STATE + myokit.LOG_BOUND,
            self._modelf,
            dims=self._ncellsf,
            global_vars=lf)
        lt = myokit.prepare_log(
            myokit.LOG_STATE + myokit.LOG_BOUND,
            self._modelt,
            dims=self._ncellst,
            global_vars=lt)
        for key in lf:
            if key not in logf:
                raise myokit.FindNanError(
                    'Method requires a simulation log from the fiber model'
                    ' containing all states and bound variables. Missing'
                    ' variable <' + key + '>.')
        for key in lt:
            if key not in logt:
                raise myokit.FindNanError(
                    'Method requires a simulation log from the tissue model'
                    ' containing all states and bound variables. Missing at'
                    ' least variable <' + key + '>.')
        del lf
        del lt

        # Error criterium: NaN/inf detection
        def bisect(ar, lo, hi):
            if not np.isfinite(ar[lo]):
                return lo
            md = lo + int(np.ceil(0.5 * (hi - lo)))
            if md == hi:
                return hi
            if not np.isfinite(ar[md]):
                return bisect(ar, lo, md)
            else:
                return bisect(ar, md, hi)

        # Search for first occurrence of propagating NaN in the log
        def find_error_position(log):
            ifirst = None   # Log list index
            kfirst = None   # Log key
            for key, ar in log.items():
                if ifirst is None:
                    if not np.isfinite(ar[-1]):
                        # First NaN found
                        kfirst = key
                        ifirst = bisect(ar, 0, len(ar) - 1)
                        if ifirst == 0:
                            break
                elif not np.isfinite(ar[ifirst - 1]):
                    # Earlier NaN found
                    kfirst = key
                    ifirst = bisect(ar, 0, ifirst)
                    if ifirst == 0:  # pragma: no cover
                        break
            return ifirst, kfirst

        # Get the name of a time variable in the fiber model
        time_varf = self._modelf.time().qname()

        # Deep searching function
        def relog(_logf, _logt, _dt):
            # Get first occurence of error
            ifirstf, kfirstf = find_error_position(_logf)
            ifirstt, kfirstt = find_error_position(_logt)
            if kfirstf is None and kfirstt is None:
                raise myokit.FindNanError('Error condition not found in logs.')
            elif kfirstf is None:  # pragma: no cover
                ifirst = ifirstt
            elif kfirstt is None:  # pragma: no cover
                ifirst = ifirstf
            elif ifirstf == 0 or ifirstt == 0:
                raise myokit.FindNanError(
                    'Unable to work with simulation logs where the error'
                    ' condition is met in the very first data point.')
            else:
                ifirst = min(ifirstf, ifirstt)
            # Position to start deep search at
            istart = ifirst - 1

            # Get last logged states before error
            statef = []
            statet = []
            for dims in myokit._dimco(*self._ncellsf):
                pre = '.'.join([str(x) for x in dims]) + '.'
                for s in self._modelf.states():
                    statef.append(_logf[pre + s.qname()][istart])
            for dims in myokit._dimco(*self._ncellst):
                pre = '.'.join([str(x) for x in dims]) + '.'
                for s in self._modelt.states():
                    statet.append(_logt[pre + s.qname()][istart])

            # Get last time before error
            time = _logf[time_varf][istart]

            # Save current state & time
            old_statef = self._statef
            old_statet = self._statet
            old_time = self._time
            self._statef = statef
            self._statet = statet
            self._time = time

            # Run until next time point, log every step
            duration = _logf[time_varf][ifirst] - time
            log = myokit.LOG_BOUND + myokit.LOG_STATE
            _logf, _logt = self.run(
                duration, logf=log, logt=log, log_interval=_dt,
                report_nan=False)

            # Reset simulation to original state
            self._statef = old_statef
            self._statet = old_statet
            self._time = old_time

            # Return new logs
            return _logf, _logt

        # Get time step
        try:
            dt = logf[time_varf][1] - logf[time_varf][0]
        except IndexError:  # pragma: no cover
            # Assume huge dt
            dt = 5

        # Search with successively fine log interval
        while dt > 0:
            dt *= 0.1
            if dt < 0.5:
                dt = 0
            logf, logt = relog(logf, logt, dt)

        # Search for first occurrence of error in the detailed log
        ifirstf, kfirstf = find_error_position(logf)
        ifirstt, kfirstt = find_error_position(logt)
        if kfirstt is None or (kfirstf is not None and kfirstf < kfirstt):
            part = 'fiber'
            ifirst = ifirstf
            kfirst = kfirstf
            if ifirst is None:  # pragma: no cover
                ifirst = 0
                kfirst = next(iter(log.items()))[0]
            model = self._modelf
            log = logf
        else:
            part = 'tissue'
            ifirst = ifirstt
            kfirst = kfirstt
            model = self._modelt
            log = logt

        # Get indices of cell in state vector
        icell = [int(x) for x in kfirst.split('.')[0:2]]

        # Get state & bound before, during and after error
        def state(index, icell):
            s = []
            b = {}
            for var in model.states():
                s.append(log[var.qname(), icell][index])
            for var in model.variables(bound=True):
                if var.binding() in self._global:
                    b[var.qname()] = log[var.qname()][index]
                else:
                    b[var.qname()] = log[var.qname(), icell][index]
            return s, b

        # Get error cell's states before, during and after
        states = []
        bound = []
        max_states = 3
        for k in range(ifirst, ifirst - max_states - 1, -1):
            if k < 0:  # pragma: no cover
                break
            s, b = state(k, icell)
            states.append(s)
            bound.append(b)

        # Get variable causing error
        var = model.get('.'.join(kfirst.split('.')[2:]))

        # Get value causing error
        if var.is_state():
            value = states[1 if ifirst > 0 else 0][var.index()]
        else:
            value = bound[1 if ifirst > 0 else 0][var.qname()]
        var = var.qname()

        # Get time error occurred
        time = logf[time_varf][ifirst]

        # Return part, time, icell, variable, value, states, bound
        return part, time, icell, var, value, states, bound

    def pre(self, duration, report_nan=True, progress=None,
            msg='Pre-pacing FiberTissueSimulation'):
        """
        This method can be used to perform an unlogged simulation, typically to
        pre-pace to a (semi-)stable orbit.

        After running this method

        - The simulation time is **not** affected
        - The current state and the default state are updated to the final
          state reached in the simulation.

        simulation to this new default state.

        Calls to :meth:`reset` after using :meth:`pre` will revert the
        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        self._run(duration, myokit.LOG_NONE, myokit.LOG_NONE, 1, report_nan,
                  progress, msg)
        self._default_statef = list(self._statef)
        self._default_statet = list(self._statet)

    def reset(self):
        """
        Resets the simulation:

        - The time variable is set to 0
        - The current state is set to the default state (either the model's
          initial state or the last state reached using :meth:`pre`)

        """
        self._time = 0
        self._statef = list(self._default_statef)
        self._statet = list(self._default_statet)

    def run(
            self, duration, logf=None, logt=None, log_interval=1.0,
            report_nan=True, progress=None,
            msg='Running FiberTissueSimulation'):
        """
        Runs a simulation and returns the logged results as a tuple containing
        two :class:`myokit.DataLog` objects.. Running a simulation has
        the following effects:

        - The internal state is updated to the last state in the simulation.
        - The simulation's time variable is updated to reflect the time
          elapsed during the simulation.

        The number of time units to simulate can be set with ``duration``.

        The variables to log can be indicated using the arguments ``logf`` and
        ``logt``. There are several options for their values:

        - ``None`` (default), to log all states
        - An integer flag or a combination of flags. Options:
          ``myokit.LOG_NONE``, ``myokit.LOG_STATE``, ``myokit.LOG_BOUND``,
          ``myokit.LOG_INTER``, ``myokit.LOG_ALL``.
        - A list of qnames or variable objects
        - A :class:`myokit.DataLog`. In this case, new data will be appended
          to the existing log.

        For more details on the log arguments, see the function
        :meth:`myokit.prepare_log`.

        Any variables bound to "time" or "pace" will be logged globally, all
        others will be logged per cell. These variables will be prefixed with a
        single number indicating the cell index in the fiber, and with two
        numbers indicating the index in the tissue.

        A log entry will be made every time *at least* ``log_interval`` time
        units have passed. No guarantee is given about the exact time log
        entries will be made, but the value of any logged time variable is
        guaranteed to be accurate.

        Intermediary variables can be logged, but with one small drawback: for
        performance reasons the logged values of states and bound variables
        will always be one time step ``dt`` ahead of the intermediary
        variables. For example if running the simulation with a step size
        ``dt=0.001`` the entry for a current ``IKr`` stored at ``t=1`` will be
        ``IKr(0.999)``, while the entry for state ``V`` will be ``V(1)``. If
        exact intermediary variables are needed it's best to log only states
        and bound variables and re-calculate the intermediary variables from
        these manually.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        r = self._run(
            duration, logf, logt, log_interval, report_nan, progress, msg)
        self._time += duration
        return r

    def _run(self, duration, logf, logt, log_interval, report_nan, progress,
             msg):
        # Simulation times
        if duration < 0:
            raise ValueError('Simulation duration can\'t be negative.')
        tmin = self._time
        tmax = tmin + duration

        # Gather global variables in fiber model
        gf = []
        gt = []
        for label in self._global:
            v = self._modelf.binding(label)
            if v is not None:
                gf.append(v.qname())
            v = self._modelt.binding(label)
            if v is not None:
                gt.append(v.qname())

        # Parse logf argument
        logf = myokit.prepare_log(
            logf,
            self._modelf,
            dims=self._ncellsf,
            global_vars=gf,
            if_empty=myokit.LOG_STATE + myokit.LOG_BOUND,
            allowed_classes=myokit.LOG_STATE + myokit.LOG_BOUND
            + myokit.LOG_INTER,
            precision=self._precision)

        # Parse logt argument
        logt = myokit.prepare_log(
            logt,
            self._modelt,
            dims=self._ncellst,
            global_vars=gt,
            if_empty=myokit.LOG_STATE + myokit.LOG_BOUND,
            allowed_classes=myokit.LOG_STATE + myokit.LOG_BOUND
            + myokit.LOG_INTER,
            precision=self._precision)

        # Create list of intermediary fiber variables that need to be logged
        inter_logf = []
        vars_checked = set()
        for var in logf.keys():
            var = myokit.split_key(var)[1]
            if var in vars_checked:
                continue
            vars_checked.add(var)
            var = self._modelf.get(var)
            if var.is_intermediary() and not var.is_bound():
                inter_logf.append(var)

        # Create list of intermediary tissue variables that need to be logged
        inter_logt = []
        vars_checked = set()
        for var in logt.keys():
            var = myokit.split_key(var)[1]
            if var in vars_checked:
                continue
            vars_checked.add(var)
            var = self._modelt.get(var)
            if var.is_intermediary() and not var.is_bound():
                inter_logt.append(var)

        # Get preferred platform/device combo from configuration file
        platform, device = myokit.OpenCL.load_selection_bytes()

        # Generate kernels
        kernel_file = os.path.join(myokit.DIR_CFUNC, KERNEL_FILE)
        args = {
            'precision': self._precision,
            'native_math': self._native_math,
            'diffusion': True,
            'fields': [],
            'rl_states': {},
            'connections': False,
            'heterogeneous': False,
        }
        args['model'] = self._modelf
        args['vmvar'] = self._vmf
        args['bound_variables'] = self._bound_variablesf
        args['inter_log'] = inter_logf
        args['paced_cells'] = self._paced_cells
        args['fiber_tissue'] = True
        kernelf = self._export(kernel_file, args, continue_in_debug_mode=True)

        args['model'] = self._modelt
        args['vmvar'] = self._vmt
        args['bound_variables'] = self._bound_variablest
        args['inter_log'] = inter_logt
        args['paced_cells'] = []
        args['fiber_tissue'] = False
        kernelt = self._export(kernel_file, args)

        # Logging period (0 = disabled)
        log_interval = 1e-9 if log_interval is None else float(log_interval)
        if log_interval <= 0:
            log_interval = 1e-9

        # Get progress indication function (if any)
        if progress is None:
            progress = myokit._Simulation_progress
        if progress:
            if not isinstance(progress, myokit.ProgressReporter):
                raise ValueError(
                    'The argument "progress" must be either a subclass of'
                    ' myokit.ProgressReporter or None.')

        # Run simulation
        if duration > 0:
            # Initialize
            state_inf = self._statef
            state_int = self._statet
            state_outf = list(state_inf)
            state_outt = list(state_int)
            self._sim.sim_init(
                platform,
                device,
                kernelf,
                kernelt,
                self._ncellsf[0],
                self._ncellsf[1],
                self._ncellst[0],
                self._ncellst[1],
                self._vmf.index(),
                self._vmt.index(),
                self._gf[0],
                self._gf[1],
                self._gt[0],
                self._gt[1],
                self._gft,
                self._cfx,
                self._ctx,
                self._cty,
                tmin,
                tmax,
                self._step_size,
                state_inf,
                state_int,
                state_outf,
                state_outt,
                self._protocol,
                logf,
                logt,
                log_interval,
                [x.qname().encode('ascii') for x in inter_logf],
                [x.qname().encode('ascii') for x in inter_logt],
            )
            try:
                t = tmin
                if progress:
                    # Loop with feedback
                    with progress.job(msg):
                        r = 1.0 / duration if duration != 0 else 1
                        while t < tmax:
                            t = self._sim.sim_step()
                            if t < tmin:    # pragma: no cover
                                # A numerical error has occurred.
                                break
                            if not progress.update(min((t - tmin) * r, 1)):
                                raise myokit.SimulationCancelledError()
                else:
                    # Loop without feedback
                    while t < tmax:
                        t = self._sim.sim_step()
                        if t < tmin:
                            # A numerical error has occurred.
                            break
            finally:
                # Clean even after KeyboardInterrupt or other Exception
                self._sim.sim_clean()
            # Update states
            self._statef = state_outf
            self._statet = state_outt

        # Check for NaN's, print error output
        if (report_nan and (logf.has_nan() or logt.has_nan())):
            txt = ['Numerical error found in simulation logs.']
            try:
                # NaN encountered, show how it happened
                part, time, icell, var, value, states, bounds = self.find_nan(
                    logf, logt)
                model = self._modelt if part == 'tissue' else self._modelf
                txt.append(
                    'Encountered numerical error in ' + part
                    + ' simulation at t ='
                    + myokit.float.str(time, precision=self._precision)
                    + ' in cell (' + ','.join([str(x) for x in icell])
                    + ') when ' + var + '= '
                    + myokit.float.str(value, precision=self._precision) + '.')

                # Get names of vars used in bindings
                vtime = model.time().qname()
                vdiff = model.binding('diffusion_current').qname()
                vpace = model.binding('pace')
                vpace = None if vpace is None else vpace.qname()

                # Show final state and bound variables
                txt.append('State during:')
                txt.append(model.format_state(
                    states[0], precision=self._precision))

                txt.append('Simulation variables during:')
                txt.append('  Time:' + myokit.float.str(
                    bounds[0][vtime], precision=self._precision))
                if vpace is not None:
                    txt.append('  Pacing variable:' + myokit.float.str(
                        bounds[0][vpace], precision=self._precision))
                txt.append('  Diffusion current:' + myokit.float.str(
                    bounds[0][vdiff], precision=self._precision))

                # Show previous state and derivatives
                n_states = len(states)
                txt.append('Obtained ' + str(n_states) + ' previous state(s).')
                if n_states > 1:
                    # Get state and bounds at previous state
                    state = states[1]
                    bound = {
                        'time': bounds[1][vtime],
                        'pace': 0 if vpace is None else bounds[1][vpace],
                        'diffusion_current': bounds[1][vdiff],
                    }

                    # Evaluate state derivatives
                    eval_error = None
                    try:
                        derivs = model.evaluate_derivatives(
                            state, bound, self._precision, ignore_errors=False)
                    except myokit.NumericalError as ee:  # pragma: no cover
                        derivs = model.evaluate_derivatives(
                            state, bound, self._precision, ignore_errors=True)
                        eval_error = str(ee).strip()

                    # Show state and derviatives
                    txt.append('State before:')
                    txt.append(model.format_state_derivatives(
                        states[1], derivs, self._precision))

                    # Show bound variables
                    txt.append('Simulation variables during:')
                    txt.append('  Time:' + myokit.float.str(
                        bounds[1][vtime], precision=self._precision))
                    if vpace is not None:
                        txt.append('  Pacing variable:' + myokit.float.str(
                            bounds[0][vpace], precision=self._precision))
                    txt.append('  Diffusion current:' + myokit.float.str(
                        bounds[1][vdiff], precision=self._precision))

                    # Show any error in evaluating derivatives
                    if eval_error is not None:  # pragma: no cover
                        txt.append('Error when evaluating derivatives:')
                        txt.append(eval_error)

            except myokit.FindNanError as e:
                txt.append(
                    'Unable to pinpoint source of NaN, an error occurred:')
                txt.append(str(e))
            raise myokit.SimulationError('\n'.join(txt))

        # Return logs
        return logf, logt

    def _set_statef(self, state, x, y, update):
        """
        Handles set_state and set_default_state for the tissue model.
        """
        n = len(state)
        ntotalf = self._ncellsf[0] * self._ncellsf[1]
        if n == self._nstatef * ntotalf:
            return list(state)
        elif n != self._nstatef:
            raise ValueError('Given state must have the same size as a single'
                             ' fiber cell state or a full simulation state')
        if x is None:
            return list(state) * ntotalf
        elif y is None:
            raise ValueError(
                'Must provide no coordinates, or both x and y coordinates.')

        # Set specific cell state
        x = int(x)
        if x < 0 or x >= self._ncellsf[0]:
            raise IndexError('Given x-index out of range.')
        y = int(y)
        if y < 0 or y >= self._ncellsf[1]:
            raise IndexError('Given y-index out of range.')
        offset = (x + y * self._ncellsf[0]) * self._nstatef
        update[offset:offset + self._nstatef] = state
        return update

    def _set_statet(self, state, x, y, update):
        """
        Handles set_state and set_default_state for the tissue model.
        """
        n = len(state)
        ntotalt = self._ncellst[0] * self._ncellst[1]
        if n == self._nstatet * ntotalt:
            return list(state)
        elif n != self._nstatet:
            raise ValueError('Given state must have the same size as a single'
                             ' tissue cell state or a full simulation state')
        if x is None:
            return list(state) * ntotalt
        elif y is None:
            raise ValueError(
                'Must provide no coordinates, or both x and y coordinates.')

        # Set specific cell state
        x = int(x)
        if x < 0 or x >= self._ncellst[0]:
            raise IndexError('Given x-index out of range.')
        y = int(y)
        if y < 0 or y >= self._ncellst[1]:
            raise IndexError('Given y-index out of range.')
        offset = (x + y * self._ncellst[0]) * self._nstatet
        update[offset: offset + self._nstatet] = state
        return update

    def set_default_fiber_state(self, state, x=None, y=None):
        """
        Changes this simulation's default state for the fiber model.

        This can be used in three different ways:

        1. When called with an argument ``state`` of size ``n_states`` and
           ``x=None`` the given state will be set as the new default state of
           all fiber cells in the simulation.
        2. Called with an argument ``state`` of size ``n_states`` and
           ``x, y`` equal to a valid cell index, this method will update only
           the selected fiber cell's default state.
        3. Finally, when called with a ``state`` of size
           ``n_states * n_total_cells``, the method will treat ``state`` as a
           concatenation of default state vectors for each fiber cell. For 2d
           simulations, the list should be indexed so that x changes first (the
           first ``nx`` entries have ``y = 0``, the second ``nx`` entries have
           ``y = 1`` and so on).

        """
        self._default_statef = self._set_statef(
            state, x, y, self._default_statef)

    def set_default_tissue_state(self, state, x=None, y=None):
        """
        Changes this simulation's default state for the tissue model.

        This can be used in three different ways:

        1. When called with an argument ``state`` of size ``n_states`` and
           ``x=None`` the given state will be set as the new default state of
           all tissue cells in the simulation.
        2. Called with an argument ``state`` of size ``n_states`` and
           ``x, y`` equal to a valid cell index, this method will update only
           the selected tissue cell's default state.
        3. Finally, when called with a ``state`` of size
           ``n_states * n_total_cells``, the method will treat ``state`` as a
           concatenation of default state vectors for each tissue cell. For 2d
           simulations, the list should be indexed so that x changes first (the
           first ``nx`` entries have ``y = 0``, the second ``nx`` entries have
           ``y = 1`` and so on).

        """
        self._default_statet = self._set_statet(
            state, x, y, self._default_statet)

    def set_fiber_state(self, state, x=None, y=None):
        """
        Changes the state of this simulation's fiber model.

        This can be used in three different ways:

        1. When called with an argument ``state`` of size ``n_states`` and
           ``x=None`` the given state will be set as the new state of all
           fiber cells in the simulation.
        2. Called with an argument ``state`` of size ``n_states`` and
           ``x, y`` equal to a valid cell index, this method will update only
           the selected fiber cell's state.
        3. Finally, when called with a ``state`` of size
           ``n_states * n_total_cells``, the method will treat ``state`` as a
           concatenation of state vectors for each fiber cell. For 2d
           simulations, the list should be indexed so that x changes first (the
           first ``nx`` entries have ``y = 0``, the second ``nx`` entries have
           ``y = 1`` and so on).

        """
        self._statef = self._set_statef(state, x, y, self._statef)

    def set_tissue_state(self, state, x=None, y=None):
        """
        Changes the state of this simulation's tissue model.

        This can be used in three different ways:

        1. When called with an argument ``state`` of size ``n_states`` and
           ``x=None`` the given state will be set as the new state of all
           tissue cells in the simulation.
        2. Called with an argument ``state`` of size ``n_states`` and
           ``x, y`` equal to a valid cell index, this method will update only
           the selected tissue cell's state.
        3. Finally, when called with a ``state`` of size
           ``n_states * n_total_cells``, the method will treat ``state`` as a
           concatenation of state vectors for each tissue cell. For 2d
           simulations, the list should be indexed so that x changes first (the
           first ``nx`` entries have ``y = 0``, the second ``nx`` entries have
           ``y = 1`` and so on).

        """
        self._statet = self._set_statet(state, x, y, self._statet)

    def set_step_size(self, step_size=0.005):
        """ Sets the solver step size. """
        step_size = float(step_size)
        if step_size <= 0:
            raise ValueError('Step size must be greater than zero.')
        self._step_size = step_size

    def set_protocol(self, protocol=None):
        """ Changes the pacing protocol used by this simulation. """
        if protocol is None:
            self._protocol = None
        else:
            self._protocol = protocol.clone()

    def set_time(self, time=0):
        """ Sets the current simulation time. """
        self._time = float(time)

    def step_size(self, step_size=0.005):
        """ Returns the solver step size. """
        return self._step_size

    def time(self):
        """ Returns the current simulation time """
        return self._time

    def tissue_shape(self):
        """
        Returns the shape of this simulation's grid of tissue cells as a tuple
        ``(ny, nx)``.
        """
        # TODO: See https://github.com/MichaelClerx/myokit/issues/764
        return (self._ncellst[1], self._ncellst[0])

    def tissue_state(self, x=None, y=None):
        """
        Returns the current simulation state in the tissue as a list of
        ``len(state_tissue) * ncells_tissue`` floating point values.

        If the optional arguments ``x`` and ``y`` specify a valid cell index a
        single cell's state is returned.
        """
        if x is None:
            return list(self._statet)
        elif y is None:
            raise ValueError('tissue_state() must be called with no arguments'
                             ' or with both an x and y argument.')
        else:
            x = int(x)
            if x < 0 or x >= self._ncellst[0]:
                raise IndexError('Given x-index out of range.')
            y = int(y)
            if y < 0 or y >= self._ncellst[1]:
                raise IndexError('Given y-index out of range.')
            x += y * self._ncellst[0]
            return self._statet[x * self._nstatet:(x + 1) * self._nstatet]




#
# Provides support for working with data in formats used by WinWCP
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._wcp import WcpFile  # noqa
# Importers
# Exporters
# Expression writers



#
# Imports selected types of protocols from files in Axon Binary Format
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit.formats


class AbfImporter(myokit.formats.Importer):
    """
    This :class:`Importer <myokit.formats.Importer>` can import protocols from
    files in Axon Binary Format.
    """

    def supports_protocol(self):
        return True

    def protocol(self, filename, channel=0):
        """
        Attempts to load the protocol from the file at ``filename``.

        If specified, the channel index ``channel`` will be used to select
        which channel in the AbfFile to convert to a protocol
        """
        from myokit.formats.axon import AbfFile
        abf = AbfFile(filename)
        return abf.da_protocol(channel)



#
# Myokit auxillary functions: This module can be used to gather any
# functions that are important enough to warrant inclusion in the main
# myokit module but don't belong to any specific hidden module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

import sys

# Globally shared numpy expression writer
_numpywriter_ = None

# Globally shared python expression writer
_pywriter_ = None


class Benchmarker(myokit.tools.Benchmarker):
    """Deprecated alias of :class:`myokit.tools.Benchmarker`."""

    def __init__(self):
        # Deprecated since 2021-03-24
        import warnings
        warnings.warn(
            'The class `myokit.Benchmarker` is deprecated.'
            ' Please use `myokit.tools.Benchmarker` instead.')
        super().__init__()


def date():
    """
    Returns the current date and time, in the format used throughout Myokit.
    """
    import time
    return time.strftime(myokit.DATE_FORMAT)


def default_protocol(model=None):
    """
    Returns a default protocol to use when no embedded one is available.
    """
    start = 100
    duration = 0.5
    period = 1000

    # Try to get the time units
    time_units = None
    if model is not None:
        if model.time() is not None:
            time_units = model.time().unit()

    # Adapt protocol if necessary
    default_units = myokit.units.ms
    try:
        start = myokit.Unit.convert(start, default_units, time_units)
        duration = myokit.Unit.convert(duration, default_units, time_units)
        period = myokit.Unit.convert(period, default_units, time_units)
    except myokit.IncompatibleUnitError:
        pass

    # Create and return
    p = myokit.Protocol()
    p.schedule(1, start, duration, period, 0)
    return p


def default_script(model=None):
    """
    Returns a default script to use when no embedded script is available.
    """
    # Defaults
    vm = 'next(m.states()).qname()'
    duration = 1000

    # Try to improve on defaults using model
    if model is not None:
        # Guess membrane potential
        import myokit.lib.guess
        v = myokit.lib.guess.membrane_potential(model)
        if v is not None:
            vm = "'" + v.qname() + "'"

        # Get duration in good units
        default_unit = myokit.units.ms
        time = model.time()
        if time is not None:
            if time.unit() != default_unit:
                try:
                    duration = myokit.Unit.convert(
                        1000, default_unit, time.unit())
                except myokit.IncompatibleUnitError:
                    pass

    # Create and return script
    return '\n'.join((  # pragma: no cover
        "[[script]]",
        "import matplotlib.pyplot as plt",
        "import myokit",
        "",
        "# Get model and protocol, create simulation",
        "m = get_model()",
        "p = get_protocol()",
        "s = myokit.Simulation(m, p)",
        "",
        "# Run simulation",
        "d = s.run(" + str(duration) + ")",
        "",
        "# Display the results",
        "var = " + vm,
        "plt.figure()",
        "plt.plot(d.time(), d[var])",
        "plt.title(var)",
        "plt.show()",
    ))


def format_float_dict(d):
    """
    Takes a dictionary of ``string : float`` mappings and returns a single
    multi-line string, where each line is of the form ``key = value``, where
    key is padded with spaces to make the equals signs align.

    This method is deprecated and will be removed in future versions of Myokit.
    """
    # Deprecated since 2021-03-24
    import warnings
    warnings.warn(
        'The method `myokit.format_float_dict` is deprecated.'
        ' It will be removed in future versions of Myokit.')

    keys = [str(k) for k in d.keys()]
    keys.sort()
    n = max([len(k) for k in keys])
    return '\n'.join([
        k + ' ' * (n - len(k)) + ' = ' + myokit.float.str(d[k]) for k in keys])


def format_path(path, root='.'):
    """Deprecated alias of :class:`myokit.tools.format_path`."""
    # Deprecated since 2021-03-24
    import warnings
    warnings.warn(
        'The method `myokit.format_path` is deprecated.'
        ' Please use `myokit.tools.format_path` instead.')
    return myokit.tools.format_path(path, root)


class ModelComparison:
    """
    Compares two models.

    The resulting diff text can be iterated over::

        for line in ModelComparison(m1, m2):
            print(line)

    Differences can be printed directly to ``stdout`` by setting ``live=True``.
    """
    def __init__(self, model1, model2, live=False):
        # Difference list
        self._diff = []

        # Live reporting
        self._live = True if live else False

        # Compare models
        if live:
            print('Comparing:')
            print('  [1] ' + model1.name())
            print('  [2] ' + model2.name())

        # -> Model meta data
        self._meta(model1, model2)

        # -> User functions
        self._userfunc(model1, model2)

        # -> Time variable
        self._time(model1, model2)

        # -> State vector
        self._state(model1, model2)

        # -> Components & Variables
        self._components(model1, model2)

        # Final report
        if live:
            print('Done')
            print('  ' + str(len(self._diff)) + ' differences found')

    def _components(self, m1, m2):
        """
        Compares two models' components.
        """
        seen = set()
        for c1 in m1.components(sort=True):
            seen.add(c1.qname())
            try:
                c2 = m2[c1.qname()]
            except KeyError:
                self._write('[2] Missing Component <' + c1.qname() + '>')
                continue
            self._comp(c1, c2)
        for c2 in m2.components(sort=True):
            if c2.qname() not in seen:
                self._write('[1] Missing Component <' + c2.qname() + '>')

    def _comp(self, c1, c2):
        """
        Compares the contents of two components with the same name.
        """
        assert c1.qname() == c2.qname()
        # Meta information
        self._meta(c1, c2)
        # Test variables
        seen = set()
        for v1 in c1.variables(deep=True):
            name = v1.qname()
            seen.add(name)
            try:
                v2 = c2.model().get(name)
            except KeyError:
                self._write('[2] Missing Variable <' + name + '>')
                continue
            self._var(v1, v2)
        for v2 in c2.variables(deep=True):
            name = v2.qname()
            if name not in seen:
                self._write('[1] Missing Variable <' + name + '>')

    def equal(self):
        """
        Returns ``True`` if the two models were equal.
        """
        return len(self._diff) == 0

    def __iter__(self):
        """
        Iterate over the found differences.
        """
        return iter(self._diff)

    def __len__(self):
        """
        Returns the length of the difference array.
        """
        return len(self._diff)

    def _meta(self, x1, x2):
        """
        Compares two objects' meta properties.
        """
        assert type(x1) == type(x2)
        if isinstance(x1, myokit.Model):
            name = ' in model'
        else:
            name = ' in <' + x1.qname() + '>'
        seen = set()
        for key, value in x1.meta.items():
            seen.add(key)
            try:
                if value != x2.meta[key]:
                    self._write(
                        '[x] Mismatched Meta property' + name + ': "'
                        + str(key) + '"')
            except KeyError:
                self._write(
                    '[2] Missing Meta property' + name + ': "' + str(key)
                    + '"')
        for key, value in x2.meta.items():
            if key not in seen:
                self._write(
                    '[1] Missing Meta property' + name + ': "' + str(key)
                    + '"')

    def _state(self, m1, m2):
        """
        Compares two models' states and initial values.
        """
        s1 = iter([v.qname() for v in m1.states()])
        s2 = iter([v.qname() for v in m2.states()])
        c1 = m1.initial_values()
        c2 = m2.initial_values()
        for k, v1 in enumerate(s1):
            try:
                v2 = next(s2)
            except StopIteration:
                self._write('[2] Missing state at position ' + str(k))
                continue
            if v1 != v2:
                self._write(
                    '[x] Mismatched State at position ' + str(k) + ': [1]<'
                    + v1 + '> [2]<' + v2 + '>')
                continue
            if c1[k].code() != c2[k].code():
                self._write('[x] Mismatched Initial value for <' + v1 + '>')
        n = m2.count_states()
        for k, v2 in enumerate(s2):
            self._write('[1] Missing state at position ' + str(n + k))

    def text(self):
        """
        Returns the full difference text.
        """
        return '\n'.join(self._diff)

    def _time(self, m1, m2):
        """
        Compares two models' time variables.
        """
        t1 = m1.time()
        t2 = m2.time()
        if t1 is None:
            if t2 is not None:
                self._write('[1] Missing Time variable <' + t2.qname() + '>')
                return
        elif t2 is None:
            self._write('[2] Missing Time variable <' + t1.qname() + '>')
            return
        if t1.qname() != t2.qname():
            self._write(
                '[x] Mismatched Time variable: [1]<' + t1.qname() + '> [2]<'
                + t2.qname() + '>')

    def _userfunc(self, m1, m2):
        """
        Compares two models' user functions.
        """
        u1 = m1.user_functions()
        u2 = m2.user_functions()
        seen = set()
        for name, func in u1.items():
            seen.add(name)
            try:
                if func != u2[name]:
                    self._write('[x] Mismatched User function <' + name + '>')
            except KeyError:
                self._write('[2] Missing User function <' + name + '>')
        for name, func in u2.items():
            if name not in seen:
                self._write('[1] Missing User function <' + name + '>.')

    def _var(self, v1, v2):
        """
        Compares two variables with the same name.
        """
        name = v1.qname()
        assert v2.qname() == name
        # Left-hand side expression
        c1, c2 = v1.lhs().code(), v2.lhs().code()
        if c1 != c2:
            self._write('[x] Mismatched LHS <' + name + '>')
        # Right-hand side expression
        c1, c2 = v1.rhs().code(), v2.rhs().code()
        if c1 != c2:
            self._write('[x] Mismatched RHS <' + name + '>')
        # Units
        if v1.unit() != v2.unit():
            self._write('[x] Mismatched unit <' + name + '>')
        # Meta
        self._meta(v1, v2)
        # Don't test nested variables, this is done component wise

    def _write(self, text):
        """
        Writes a line to the difference array.
        """
        if self._live:
            print(text)
        self._diff.append(text)


def numpy_writer():
    """
    Returns a globally shared numpy expression writer.

    LhsExpressions are converted as follows:

    1. Derivatives are indicated as "_d_" + var.uname()
    2. Other names are indicated as var.uname()

    This convention ensures a unique mapping of a model's lhs expressions to
    acceptable python variable names.
    """
    global _numpywriter_
    if _numpywriter_ is None:
        from .formats.python import NumPyExpressionWriter
        _numpywriter_ = NumPyExpressionWriter()

        def name(x):
            u = x.var()
            u = u.uname() if isinstance(u, myokit.ModelPart) else str(u)
            if u is None:
                u = x.var().qname().replace('.', '_')
            return '_d_' + u if x.is_derivative() else u

        _numpywriter_.set_lhs_function(name)

    return _numpywriter_


def python_writer():
    """
    Returns a globally shared python expression writer.

    LhsExpressions are converted as follows:

    1. Derivatives are indicated as "_d_" + var.uname()
    2. Other names are indicated as var.uname()

    This convention ensures a unique mapping of a model's lhs expressions to
    acceptable python variable names.
    """
    global _pywriter_
    if _pywriter_ is None:
        from .formats.python import PythonExpressionWriter
        _pywriter_ = PythonExpressionWriter()

        def name(x):
            u = x.var()
            u = u.uname() if isinstance(u, myokit.ModelPart) else str(u)
            if u is None:
                u = x.var().qname().replace('.', '_')
            return '_d_' + u if x.is_derivative() else u

        _pywriter_.set_lhs_function(name)

    return _pywriter_


def _prepare_bindings(model, labels):
    """
    Takes a mapping of binding labels to internal references and returns a map
    from variable instances to the same internal references.

    This method also modifies ``model``: All bindings that are not mapped to
    an internal reference will be removed.

    The argument ``mapping`` should take the form::

        labels = {
            binding_label_1 : internal_name_1,
            binding_label_2 : internal_name_2,
            ...
            }

    The returned dictionary will have the form::

        variables = {
            variable_x : internal_name_1,
            variable_y : internal_name_2,
            ...
            }

    Unsupported bindings (i.e. bindings not appearing in ``labels``) will
    be ignored.
    """
    unused = []
    variables = {}
    for label, var in model.bindings():
        try:
            variables[var] = labels[label]
        except KeyError:
            unused.append(var)
            continue
        # TODO: Remove this line; https://github.com/myokit/myokit/issues/320
        var.set_rhs(0)

    for var in unused:
        var.set_binding(None)

    return variables


def run(model, protocol, script, stdout=None, stderr=None, progress=None):
    """
    Runs a python ``script`` using the given ``model`` and ``protocol``.

    The following functions are provided to the script:

    ``get_model()``
        This returns the model passed to ``run()``. When using the GUI, this
        returns the model currently loaded into the editor.
    ``get_protocol()``
        This returns the protocol passed to ``run()``. When using the GUI, this
        returns the protocol currently loaded into the editor.

    Ordinary and error output can be re-routed by providing objects with a
    file-like interface (``x.write(text)``, ``x.flush()``) as ``stdout`` and
    ``stderr`` respectively. Note that output is only re-routed on the python
    level so that any output generated by C-code will still go to the main
    process's output and error streams.

    An object implementing the ``ProgressReporter`` interface can be passed in.
    This will be made available globally to any simulations (or other methods)
    that provide progress update information.
    """
    # Trim "[[script]]" from script
    script = script.splitlines()
    if script and script[0].strip() == '[[script]]':
        script = script[1:]
    script = '\n'.join(script)

    # Add an empty line at the start of the script. This will make the line
    # numbers reported in any error message correspond to those in the GUI.
    # In other words, a script section
    #
    #   1 [[script]]
    #   2 import myokit
    #   3 x = 1/0
    #
    # will raise an exception on line 3.
    #
    script = '\n' + script

    # Class to run scripts
    class Runner:
        def __init__(self, model, protocol, script, stdout, stderr, progress):
            super().__init__()
            self.model = model
            self.protocol = protocol
            self.script = script
            self.stdout = stdout
            self.stderr = stderr
            self.progress = progress
            self.exc_info = None

        def run(self):

            # Create magic functions
            def get_model():
                return self.model

            def get_protocol():
                return self.protocol

            # Re-route standard outputs, execute code
            oldstdout = oldstderr = None
            oldprogress = myokit._Simulation_progress
            myokit._Simulation_progress = self.progress
            try:
                if self.stdout is not None:
                    oldstdout = sys.stdout
                    sys.stdout = self.stdout
                if self.stderr is not None:
                    oldstderr = sys.stderr
                    sys.stderr = self.stderr
                environment = {
                    'get_model': get_model,
                    'get_protocol': get_protocol,
                }
                exec(self.script, environment)
            finally:
                if oldstdout is not None:
                    sys.stdout = oldstdout
                if oldstderr is not None:
                    sys.stderr = oldstderr
                myokit._Simulation_progress = oldprogress

    r = Runner(model, protocol, script, stdout, stderr, progress)
    r.run()

    # Free some space
    del r
    import gc
    gc.collect()


def step(model, initial=None, reference=None, ignore_errors=False):
    """
    Evaluates the state derivatives in a model and compares the results with a
    list of reference values, if given.

    The ``model`` should be given as a valid :class:`myokit.Model`. The state
    values to start from can be given as ``initial``, which should be any item
    that can be converted to a valid state using ``model.map_to_state``.

    The values can be compared to reference output given as a list
    ``reference``. Alternatively, if ``reference`` is a model the two models'
    outputs will be compared.

    By default, the evaluation routine checks for numerical errors
    (divide-by-zeros, invalid operations and overflows). To run an evaluation
    without error checking, set ``ignore_errors=True``.

    Returns a string indicating the results.
    """
    # Get initial state
    if initial is None:
        initial = model.initial_values(as_floats=True)

    # Get evaluation at initial state
    values = model.evaluate_derivatives(
        state=initial, ignore_errors=ignore_errors)

    # Log settings
    fmat = myokit.SFDOUBLE
    line_width = 79

    # Get max variable name width (at least 4, for 'Name' header)
    w = max(4, max([len(v.qname()) for v in model.states()]))

    # Create log header
    log = []
    log.append('Evaluating state vector derivatives...')
    log.append('-' * line_width)
    log.append(('{:<' + str(w) + '}  {:<24}  {:<24}').format(
        'Name', 'Initial value', 'Derivative at t=0'))
    log.append('-' * line_width)

    # Format for states
    f = '{: <' + str(w) + '}  ' + fmat + '  ' + fmat

    if not reference:
        # Default output: intial value and derivative
        for r, v in enumerate(model.states()):
            log.append(f.format(v.qname(), initial[r], values[r]))
    else:
        # Comparing output

        # Reference should be a state evaluation, or a model
        if isinstance(reference, myokit.Model):
            reference = reference.evaluate_derivatives(
                state=initial, ignore_errors=ignore_errors)

        h = ' ' * (w + 28)
        i = h + ' ' * 20
        g = h + fmat
        errors = 0
        warnings = 0

        for r, v in enumerate(model.states()):
            x = values[r]
            y = reference[r]
            log.append(f.format(v.qname(), initial[r], x))
            xx = fmat.format(x)
            yy = fmat.format(y)
            line = g.format(y)

            # Sign error
            if xx[0] != yy[0]:

                # Ignore if zero
                if (myokit.float.eq(x, 0) and myokit.float.eq(y, 0)):
                    log.append(line)
                    log.append('')
                else:
                    errors += 1
                    log.append(line + ' sign')
                    log.append(h + '^' * 24)

            # Different exponent, huge error
            elif xx[-4:] != yy[-4:]:
                errors += 1
                log.append(line + ' exponent')
                log.append(i + '^^^^')

            # Large error, small error, or no error
            else:
                mark_error = False
                abs_err = abs(x - y)
                if abs_err:
                    rel_err = abs(x - y) / max(abs(x), abs(y))
                    n_eps = rel_err / sys.float_info.epsilon
                else:
                    n_eps = rel_err = 0
                if n_eps > 1e6:
                    # "Large" error
                    errors += 1
                    line += ' X'
                    mark_error = True
                elif xx != yy:
                    # "Small" error, or numerical error
                    rel_err = abs(x - y) / max(abs(x), abs(y))
                    n_eps = rel_err / sys.float_info.epsilon
                    if n_eps > 1:
                        line += ' ~ ' + str(round(n_eps, 1)) + ' eps'
                    else:
                        line += ' <= 1 eps'
                    if n_eps > 1:
                        warnings += 1
                        mark_error = True
                log.append(line)

                if mark_error:
                    line2 = h
                    pos = 0
                    n = len(xx)
                    while pos < n and xx[pos] == yy[pos]:
                        line2 += ' '
                        pos += 1
                    for pos in range(pos, n):
                        line2 += '^'
                    log.append(line2)
                else:
                    log.append('')

        # Show large mismatches between model and reference
        if errors > 0:
            log.append(
                'Found (' + str(errors) + ') large mismatches between output'
                ' and reference values.')
        else:
            log.append('Model check completed without errors.')

        # Show small mismatches between model and reference
        if warnings > 0:
            log.append('Found (' + str(warnings) + ') small mismatches.')

    # Finalize and return
    log.append('-' * line_width)
    return '\n'.join(log)


def strfloat(number, full=False, precision=myokit.DOUBLE_PRECISION):
    """Deprecated alias of :class:`myokit.float.str`."""
    # Deprecated since 2021-03-24
    import warnings
    warnings.warn(
        'The method `myokit.strfloat` is deprecated.'
        ' Please use `myokit.float.str` instead.')
    return myokit.float.str(number, full, precision)


def time():
    """
    Returns the current time, in the format used throughout Myokit.
    """
    import time as t
    return t.strftime(myokit.TIME_FORMAT)


def version(raw=False):
    """
    Returns the current Myokit version.

    By default, a formatted multi-line string is returned. To get a simpler
    one-line string set the optional argument ``raw=True``.

    The same version info can be accessed using ``myokit.__version__``.
    Unformatted info is available in ``myokit.__version_tuple__``, which
    contains the major, minor, and revision number respectively, all as
    ``int`` objects. For development versions of Myokit, it may contain a 4th
    element ``"dev"``.
    """
    if raw:
        return myokit.__version__
    else:
        t1 = ' Myokit ' + myokit.__version__ + ' '
        t2 = '_' * len(t1)
        t1 += '|/\\'
        t2 += '|  |' + '_' * 5
        return '\n' + t1 + '\n' + t2




#
# Sympy expression reader
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit
import myokit.formats


class SymPyExpressionReader:
    """
    Converts Sympy expressions to Myokit expressions.

    Any variable names will be resolved against the given model. If no model is
    given, string names will be used instead of myokit Variable objects.
    """
    def __init__(self, model=None):
        self._model = model
        self._op_map = self._build_op_map()

    def ex(self, e):
        """
        Converts the Sympy expression ``e`` to a :class:`myokit.Expression`.
        """
        from sympy.core.numbers import Number
        try:
            action = self._op_map[type(e)]
        except KeyError:
            if isinstance(e, Number):
                action = self._ex_number
            else:
                raise ValueError('Unsupported type: ' + str(type(e)))
        return action(e)

    def set_model(self, model=None):
        """
        Changes the model used by this reader to resolve variable names.
        """
        self._model = model

    def _build_op_map(self):
        """
        Creates and returns a mapping of sympy object types to handling
        methods.
        """
        from sympy.core.symbol import Symbol
        from sympy.core.add import Add
        from sympy.core.mul import Mul
        from sympy.core.mod import Mod
        from sympy.core.power import Pow
        from sympy.functions.elementary.trigonometric import (
            sin, cos, tan, asin, acos, atan)
        from sympy.functions.elementary.exponential import exp, log
        from sympy.functions.elementary.integers import floor, ceiling
        from sympy.functions.elementary.complexes import Abs
        from sympy.logic.boolalg import Not
        from sympy.core.relational import (
            Equality, Unequality, StrictGreaterThan, StrictLessThan,
            GreaterThan, LessThan)
        from sympy.logic.boolalg import And, Or
        from sympy.functions.elementary.piecewise import Piecewise

        return {
            Symbol: self._ex_name,
            Add: self._ex_plus,
            Mul: self._ex_multiply,
            Mod: self._ex_remainder,
            Pow: self._ex_power,
            sin: self._ex_sin,
            cos: self._ex_cos,
            tan: self._ex_tan,
            asin: self._ex_asin,
            acos: self._ex_acos,
            atan: self._ex_atan,
            exp: self._ex_exp,
            log: self._ex_log,
            floor: self._ex_floor,
            ceiling: self._ex_ceil,
            Abs: self._ex_abs,
            Not: self._ex_not,
            Equality: self._ex_equal,
            Unequality: self._ex_not_equal,
            StrictGreaterThan: self._ex_more,
            StrictLessThan: self._ex_less,
            GreaterThan: self._ex_more_equal,
            LessThan: self._ex_less_equal,
            And: self._ex_and,
            Or: self._ex_or,
            Piecewise: self._ex_piecewise,
        }

    def _ex_name(self, e):
        var = str(e)
        # Check if this is a derivative
        # See :meth:`SymPyExpressionWriter._ex_derivative()`.
        if self._model:
            if var[:4] == 'dot(' and var[-1:] == ')':
                var = self._model.get(var[4:-1], myokit.Variable)
                return myokit.Derivative(myokit.Name(var))
            var = self._model.get(var, myokit.Variable)
        return myokit.Name(var)

    def _ex_number(self, e):
        return myokit.Number(float(e))

    def _ex_plus(self, e):
        a, b = e.as_two_terms()
        return myokit.Plus(self.ex(a), self.ex(b))

    def _ex_multiply(self, e):
        from sympy.core.numbers import NegativeOne
        a, b = e.as_two_terms()
        if type(a) == NegativeOne:
            return myokit.PrefixMinus(self.ex(b))
        return myokit.Multiply(self.ex(a), self.ex(b))

    def _ex_remainder(self, e):
        a, b = e.args
        return myokit.Remainder(self.ex(a), self.ex(b))

    def _ex_power(self, e):
        a, b = e.args
        return myokit.Power(self.ex(a), self.ex(b))

    def _ex_sin(self, e):
        return myokit.Sin(self.ex(e.args[0]))

    def _ex_cos(self, e):
        return myokit.Cos(self.ex(e.args[0]))

    def _ex_tan(self, e):
        return myokit.Tan(self.ex(e.args[0]))

    def _ex_asin(self, e):
        return myokit.ASin(self.ex(e.args[0]))

    def _ex_acos(self, e):
        return myokit.ACos(self.ex(e.args[0]))

    def _ex_atan(self, e):
        return myokit.ATan(self.ex(e.args[0]))

    def _ex_exp(self, e):
        return myokit.Exp(self.ex(e.args[0]))

    def _ex_log(self, e):
        return myokit.Log(*[self.ex(x) for x in e.args])

    def _ex_floor(self, e):
        return myokit.Floor(self.ex(e.args[0]))

    def _ex_ceil(self, e):
        return myokit.Ceil(self.ex(e.args[0]))

    def _ex_abs(self, e):
        return myokit.Abs(self.ex(e.args[0]))

    def _ex_not(self, e):
        return myokit.Not(self.ex(e.args[0]))

    def _ex_equal(self, e):
        return myokit.Equal(*[self.ex(x) for x in e.args])

    def _ex_not_equal(self, e):
        return myokit.NotEqual(*[self.ex(x) for x in e.args])

    def _ex_more(self, e):
        return myokit.More(*[self.ex(x) for x in e.args])

    def _ex_less(self, e):
        return myokit.Less(*[self.ex(x) for x in e.args])

    def _ex_more_equal(self, e):
        return myokit.MoreEqual(*[self.ex(x) for x in e.args])

    def _ex_less_equal(self, e):
        return myokit.LessEqual(*[self.ex(x) for x in e.args])

    def _ex_and(self, e):
        return myokit.And(*[self.ex(x) for x in e.args])

    def _ex_or(self, e):
        return myokit.Or(*[self.ex(x) for x in e.args])

    def _ex_piecewise(self, e):
        from sympy.logic.boolalg import BooleanTrue, BooleanFalse

        args = []
        n = len(e.args) - 1
        for k, pair in enumerate(e.args):
            expr, cond = pair
            expr = self.ex(expr)
            cond_is_true = False
            if type(cond) == BooleanTrue:
                cond_is_true = True
                cond = myokit.Equal(myokit.Number(1), myokit.Number(1))
            elif type(cond) == BooleanFalse:    # pragma: no cover
                # Cover pragma: SymPy currently optimises out BooleanFalse
                cond = myokit.Equal(myokit.Number(0), myokit.Number(1))
            else:
                cond = self.ex(cond)
            if k < n:
                args.append(cond)
                args.append(expr)
        if cond_is_true:
            args.append(expr)
        else:
            args.append(cond)
            args.append(expr)
            args.append(myokit.Number(0))
        return myokit.Piecewise(*args)



#
# Latex expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit.formats


class LatexExpressionWriter(myokit.formats.ExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to their Tex equivalent.
    """
    def __init__(self):
        super().__init__()

        # Default time variable
        self._time = None
        self.set_time_variable_name()

        # Default lhs function
        def fhls(lhs):
            var = self._prepare_name(lhs.var().uname())
            var = '\\text{' + var + '}'     # Depends on amsmath package!
            if isinstance(lhs, myokit.Derivative):
                var = '\\frac{d}{d\\text{t}}' + var
            return var
        self._flhs = None
        self.set_lhs_function(fhls)

    def set_lhs_function(self, f):
        """
        Sets a naming function, will be called to get the variable name from a
         ``myokit.LhsExpression`` object.

        The argument ``f`` should be a function that takes an ``LhsExpression``
        as input and returns a string.
        """
        self._flhs = f

    def set_time_variable_name(self, name='t'):
        """
        Sets a name to use for the time variable in derivatives
        """
        self._time = self._prepare_name(name)

    def eq(self, eq):
        """
        Converts an equation to a string.
        """
        return self.ex(eq.lhs) + ' = ' + self.ex(eq.rhs)

    def ex(self, e):
        """
        Converts an expression to a string.
        """
        b = []
        self._ex(e, b)
        return ''.join(b)

    def _prepare_name(self, text):
        """
        Prepares a name for use in latex
        """
        text = str(text)
        return text.replace('_', '\_')

    def _ex(self, e, b):
        try:
            action = self._op_map[type(e)]
        except KeyError:
            raise ValueError('Unknown expression type: ' + str(type(e)))
        action(e, b)

    def _ex_infix(self, e, b, op):
        """
        Handles _ex() for infix operators
        """
        if e.bracket(e[0]):
            b.append('\\left(')
        self._ex(e[0], b)
        if e.bracket(e[0]):
            b.append('\\right)')
        b.append(op)
        if e.bracket(e[1]):
            b.append('\\left(')
        self._ex(e[1], b)
        if e.bracket(e[1]):
            b.append('\\right)')

    def _ex_function(self, e, b, func):
        """
        Handles _ex() for function operators
        """
        b.append(func)
        b.append('\\left(')
        b.append(','.join([self.ex(x) for x in e]))
        b.append('\\right)')

    def _ex_infix_condition(self, e, b, op):
        """
        Handles _ex() for infix condition operators
        """
        b.append('\\left(')
        self._ex(e[0], b)
        b.append(op)
        self._ex(e[1], b)
        b.append('\\right)')

    def _ex_name(self, e, b):
        b.append(self._flhs(e))

    def _ex_derivative(self, e, b):
        b.append(self._flhs(e))

    def _ex_number(self, e, b):
        b.append(myokit.float.str(e))

    def _ex_prefix_plus(self, e, b):
        self._ex(e[0], b)

    def _ex_prefix_minus(self, e, b):
        b.append('\\left(')
        b.append('-')
        if e.bracket(e[0]):
            b.append('\\left(')
        self._ex(e[0], b)
        if e.bracket(e[0]):
            b.append('\\right)')
        b.append('\\right)')

    def _ex_plus(self, e, b):
        self._ex_infix(e, b, '+')

    def _ex_minus(self, e, b):
        self._ex_infix(e, b, '-')

    def _ex_multiply(self, e, b):
        self._ex_infix(e, b, '*')

    def _ex_divide(self, e, b):
        b.append('\\frac{')
        self._ex(e[0], b)
        b.append('}{')
        self._ex(e[1], b)
        b.append('}')

    def _ex_quotient(self, e, b):
        # Note: Quotient in myokit uses round-to-zero (like Python does!)
        # See: myokit.Quotient
        b.append('\\left\\lfloor')
        self._ex_divide(e, b)
        b.append('\\right\\rfloor')

    def _ex_remainder(self, e, b):
        self._ex_function(e, b, '\\bmod')

    def _ex_power(self, e, b):
        if e.bracket(e[0]):
            b.append('\\left(')
        self._ex(e[0], b)
        if e.bracket(e[0]):
            b.append('\\right)')
        b.append('^{')
        self._ex(e[1], b)
        b.append('}')

    def _ex_sqrt(self, e, b):
        b.append('\\sqrt{')
        self._ex(e[0], b)
        b.append('}')

    def _ex_sin(self, e, b):
        self._ex_function(e, b, '\\sin')

    def _ex_cos(self, e, b):
        self._ex_function(e, b, '\\cos')

    def _ex_tan(self, e, b):
        self._ex_function(e, b, '\\tan')

    def _ex_asin(self, e, b):
        self._ex_function(e, b, '\\arcsin')

    def _ex_acos(self, e, b):
        self._ex_function(e, b, '\\arccos')

    def _ex_atan(self, e, b):
        self._ex_function(e, b, '\\arctan')

    def _ex_exp(self, e, b):
        self._ex_function(e, b, '\\exp')

    def _ex_log(self, e, b):
        b.append('\\log')
        if len(e) > 1:
            b.append('_{')
            self._ex(e[1], b)
            b.append('}')
        b.append('\\left(')
        self._ex(e[0], b)
        b.append('\\right)')

    def _ex_log10(self, e, b):
        return self._ex_log(myokit.Log(e[0], myokit.Number(10)), b)

    def _ex_floor(self, e, b):
        b.append('\\left\\lfloor{')
        self._ex(e[0], b)
        b.append('}\\right\\rfloor')

    def _ex_ceil(self, e, b):
        b.append('\\left\\lceil{')
        self._ex(e[0], b)
        b.append('}\\right\\rceil')

    def _ex_abs(self, e, b):
        b.append('\\lvert{')
        self._ex(e[0], b)
        b.append('}\\rvert')

    def _ex_not(self, e, b):
        self._ex_function(e, b, '\\not')

    def _ex_equal(self, e, b):
        self._ex_infix_condition(e, b, '=')

    def _ex_not_equal(self, e, b):
        self._ex_infix_condition(e, b, '\\neq')

    def _ex_more(self, e, b):
        self._ex_infix_condition(e, b, '>')

    def _ex_less(self, e, b):
        self._ex_infix_condition(e, b, '<')

    def _ex_more_equal(self, e, b):
        self._ex_infix_condition(e, b, '\\geq')

    def _ex_less_equal(self, e, b):
        self._ex_infix_condition(e, b, '\\leq')

    def _ex_and(self, e, b):
        self._ex_infix_condition(e, b, '\\and')

    def _ex_or(self, e, b):
        self._ex_infix_condition(e, b, '\\or')

    def _ex_if(self, e, b):
        # Not suported
        self._ex_function(e, b, 'if')

    def _ex_piecewise(self, e, b):
        # Not suported
        self._ex_function(e, b, 'piecewise')



#
# Provides MathML (version 2) support.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._ewriter import MathMLExpressionWriter
from ._parser import (   # noqa
    MathMLError,
    MathMLParser,
    parse_mathml_etree,
    parse_mathml_string,
)

# Namespaces
NS_MATHML_2 = 'http://www.w3.org/1998/Math/MathML'

# Importers
# Exporters

# Expression writers
_ewriters = {
    'mathml': MathMLExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)

# Language keywords
# None!



#!/usr/bin/env python3
#
# Tests the OpenCL simulation classes' interpretation of log_interval
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit

from myokit.tests import OpenCL_FOUND, DIR_DATA, WarningCollector
from myokit.tests.test_simulation_log_interval import PeriodicTest

debug = False


@unittest.skipIf(not OpenCL_FOUND, 'OpenCL not found on this system.')
class SimulationOpenCLTest(PeriodicTest):
    """
    Tests myokit.SimulationOpenCL for consistent log entry timing.
    """
    def test_periodic(self):
        # Test periodic logging.

        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        s = myokit.SimulationOpenCL(m, p, ncells=1)
        self.periodic(s)


@unittest.skipIf(not OpenCL_FOUND, 'OpenCL not found on this system.')
class FiberTissueSimulationTest(unittest.TestCase):
    """
    Tests myokit.FiberTissueSimulation for consistent log entry timing.
    """
    def test_periodic(self):
        # Test periodic logging.

        m, p, x = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        with WarningCollector():
            s = myokit.FiberTissueSimulation(
                m, m, p, ncells_fiber=(1, 1), ncells_tissue=(1, 1))

        # Set tolerance for equality testing
        emax = 1e-2  # Time steps for logging are approximate

        # Test 1: Simple 5 ms simulation, log_interval 0.5 ms
        d, e = s.run(
            5, logf=['engine.time'], logt=myokit.LOG_NONE, log_interval=0.5)
        d = d.npview()
        t = d['engine.time']
        q = np.arange(0, 5, 0.5)
        if debug:
            print(t)
            print(q)
            print('- ' * 10)
        self.assertEqual(len(t), len(q))
        self.assertTrue(np.max(np.abs(t - q)) < emax)

        # Test 2: Very short simulation
        s.reset()
        d, e = s.run(
            1, logf=['engine.time'], logt=myokit.LOG_NONE, log_interval=0.5)
        d = d.npview()
        t = d['engine.time']
        q = np.arange(0, 1, 0.5)
        if debug:
            print(t)
            print(q)
            print('- ' * 10)
        self.assertEqual(len(t), len(q))
        self.assertTrue(np.max(np.abs(t - q)) < emax)

        # Test 3: Stop and start a simulation
        s.reset()
        d, e = s.run(
            1, logf=['engine.time'], logt=myokit.LOG_NONE, log_interval=0.5)
        d, e = s.run(2, logf=d, logt=myokit.LOG_NONE, log_interval=0.5)
        d, e = s.run(2, logf=d, logt=myokit.LOG_NONE, log_interval=0.5)
        d = d.npview()
        t = d['engine.time']
        q = np.arange(0, 5, 0.5)
        if debug:
            print(t)
            print(q)
            print('- ' * 10)
        self.assertEqual(len(t), len(q))
        self.assertTrue(np.max(np.abs(t - q)) < emax)


if __name__ == '__main__':
    print('Add -v for more debug output')
    import sys
    if '-v' in sys.argv:
        debug = True
    unittest.main()



#!/usr/bin/env python3
#
# Tests the Jacobian calculator.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit

from myokit.tests import DIR_DATA


class JacobianCalculatorTest(unittest.TestCase):
    """
    Tests the JacobianCalculator.
    """
    def test_simple(self):
        # Load model
        m = myokit.load_model(os.path.join(DIR_DATA, 'noble-1962.mmt'))

        # Run a simple simulation
        c = myokit.JacobianCalculator(m)
        x, f, j, e = c.newton_root(damping=0.01, max_iter=50)

        # Test if still runs with initial x all zero
        # (But does cause a linear algebra issue in this case)
        x = np.zeros(m.count_states())
        c.newton_root(x, damping=0.01, max_iter=50)

        # Test if still works with a single zero (Enno's bug)
        x = np.array(m.initial_values(True))
        x[1] = 0
        x, f, j, e = c.newton_root(x, damping=0.01, max_iter=50)

        # Test quick return
        x = np.array(m.initial_values(True))
        x[0] = 0
        x2, f, j, e = c.newton_root(damping=0.01, max_iter=1)
        self.assertTrue(np.sum((x2 - x)**2) > 10)

        # Invalid damping value
        self.assertRaisesRegex(
            ValueError, 'Damping', c.newton_root, damping=0)
        self.assertRaisesRegex(
            ValueError, 'Damping', c.newton_root, damping=1.1)

        # Missing a state
        x = m.initial_values(True)[:-1]
        self.assertRaisesRegex(
            ValueError, 'must have length', c.calculate, x)

        # Non-numbers in state
        x = m.initial_values(True)
        x[0] = 'Hello'
        self.assertRaisesRegex(
            ValueError, 'floats', c.calculate, x)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests some of the ProgressReporter classes.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest
import re
import time

import myokit


class ProgressPrinterTests(unittest.TestCase):
    """
    Tests the :class:`ProgressPrinter`.
    """
    def test_progress_printer(self):
        # Test basic functionality.

        # Zero call
        with myokit.tools.capture() as c:
            p = myokit.ProgressPrinter()
            self.assertTrue(p.update(0))
        pattern1 = re.compile(
            r'\[[0-9]{1}\.[0-9]{1} minutes\] [0-9]+(.[0-9])? % done')
        lines = c.text().splitlines()
        self.assertTrue(len(lines) > 0)
        for line in lines:
            self.assertTrue(pattern1.match(line))

        # Normal call
        with myokit.tools.capture() as c:
            p = myokit.ProgressPrinter()
            self.assertTrue(p.update(0.49))
        pattern2 = re.compile(
            r'\[[0-9]{1}\.[0-9]{1} minutes] [0-9]+(.[0-9])? % done,'
            r' estimated [0-9]{1} seconds remaining')
        lines = c.text().splitlines()
        self.assertTrue(len(lines) > 0)
        for line in lines:
            self.assertTrue(pattern2.match(line))

        # Call that will take minutes
        with myokit.tools.capture() as c:
            p = myokit.ProgressPrinter()
            time.sleep(0.1)
            self.assertTrue(p.update(1e-3))
        self.assertIn('minutes remaining', c.text())

        # Note: Printer must be created withing capture, otherwise it
        # will print to stdout (which won't have been redirected yet).

        # Status every ten percent
        with myokit.tools.capture() as c:
            p = myokit.ProgressPrinter(digits=-1)
            self.assertTrue(p.update(0))
            self.assertTrue(p.update(0.08))
            self.assertTrue(p.update(0.18))
            self.assertTrue(p.update(0.19))
            self.assertTrue(p.update(0.199))
            self.assertTrue(p.update(1))
        lines = c.text().splitlines()
        self.assertTrue(len(lines), 3)
        self.assertTrue(pattern1.match(lines[0]))
        self.assertTrue(pattern2.match(lines[1]))
        self.assertTrue(pattern2.match(lines[2]))


class TimeoutTest(unittest.TestCase):
    """
    Tests cancelling a simulation with the ``Timeout`` class.
    """
    def test_timeout_progress_reporter(self):

        m, p, x = myokit.load('example')
        s = myokit.Simulation(m, p)
        t = myokit.Timeout(0.1)
        with self.assertRaises(myokit.SimulationCancelledError):
            s.run(100000, progress=t)

        t = myokit.Timeout(3600)
        s.run(100, progress=t)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the importers for various formats
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit
import myokit.formats
import myokit.formats.channelml

from myokit.tests import DIR_FORMATS, WarningCollector

# ChannelML dir
DIR = os.path.join(DIR_FORMATS, 'channelml')


class ChannelMLTest(unittest.TestCase):
    """ Test ChannelML importing. """

    def test_capability_reporting(self):
        # Test if the right capabilities are reported.
        i = myokit.formats.importer('channelml')
        self.assertTrue(i.supports_component())
        self.assertTrue(i.supports_model())
        self.assertFalse(i.supports_protocol())

    def test_model(self):
        # Test :meth:`ChannelMLImporter.model()`.
        i = myokit.formats.importer('channelml')
        self.assertTrue(i.supports_model())
        m = i.model(os.path.join(DIR, 'ch-00-valid-file.channelml'))
        m.validate()

    def test_component(self):
        # Test :meth:`ChannelMLImporter.component()`.
        path = os.path.join(DIR, 'ch-00-valid-file.channelml')
        i = myokit.formats.importer('channelml')
        self.assertTrue(i.supports_component())
        m = myokit.Model()
        c = m.add_component('membrane')
        v = c.add_variable('V')
        self.assertRaises(myokit.ImportError, i.component, path, m)
        v.set_label('membrane_potential')
        i.component(path, m)
        cs = [c for c in m.components()]
        self.assertEqual(len(cs), 2)

    def test_error_handling(self):
        # Test various errors when loading ChannelML files.

        i = myokit.formats.importer('channelml')
        ce = myokit.formats.channelml.ChannelMLError

        # Wrong XML root element
        self.assertRaisesRegex(
            ce, 'Unknown root', i.model,
            os.path.join(DIR, 'ch-01-wrong-root.channelml'))

        # No channel_type element
        self.assertRaisesRegex(
            ce, 'No <channel_type>', i.model,
            os.path.join(DIR, 'ch-02-no-channel-type.channelml'))

        # Test name adaptation when it overlaps with names already in the model
        m = i.model(
            os.path.join(DIR, 'ch-03-overlapping-name.channelml'))
        m.validate()
        self.assertIn('membrane_2', m)  # Renamed channel membrane > membrane_2

        # No current_voltage_relation element
        self.assertRaisesRegex(
            ce, 'current voltage relation', i.model,
            os.path.join(DIR, 'ch-04-no-cvr.channelml'))

        # Multiple current_voltage_relation elements
        with WarningCollector() as w:
            m = i.model(os.path.join(DIR, 'ch-05-two-cvrs.channelml'))
        m.validate()
        self.assertEqual(w.count(), 1)
        self.assertIn('Multiple current voltage relations', w.text())

        # No q10 information
        m = i.model(os.path.join(DIR, 'ch-06-no-q10.channelml'))
        m.validate()

        # Not exactly 2 transitions
        self.assertRaisesRegex(
            ce, 'exactly 2 transitions', i.model,
            os.path.join(DIR, 'ch-07-three-transitions.channelml'))

        # Not closed-to-open transition
        self.assertRaisesRegex(
            ce, 'closed-to-open', i.model,
            os.path.join(DIR, 'ch-08-no-closed-to-open.channelml'))

        # Not open-to-closed transition
        self.assertRaisesRegex(
            ce, 'open-to-closed', i.model,
            os.path.join(DIR, 'ch-09-no-open-to-closed.channelml'))

        # Unreadable closed-to-open expression
        with WarningCollector() as w:
            m = i.model(
                os.path.join(DIR, 'ch-10-tco-bad-expression.channelml'))

        # > RHS Should not be set
        self.assertRaises(myokit.MissingRhsError, m.validate)

        # > Warning should be generated
        self.assertEqual(w.count(), 1)
        self.assertIn('expression for closed-to-open', w.text())

        # > Unparsed expression should be in meta-data
        self.assertEqual('bort', m.get('Nav1_3.m.alpha').meta['expression'])

        # Unreadable open-to-closed expression
        with WarningCollector() as w:
            m = i.model(
                os.path.join(DIR, 'ch-11-toc-bad-expression.channelml'))
        self.assertRaises(myokit.MissingRhsError, m.validate)
        self.assertEqual(w.count(), 1)
        self.assertIn('expression for open-to-closed', w.text())
        self.assertEqual('bort', m.get('Nav1_3.m.beta').meta['expression'])

        # No transitions or steady-state/time-course for a gate
        self.assertRaisesRegex(
            ce, 'steady state', i.model,
            os.path.join(DIR, 'ch-12-no-steady-state.channelml'))
        self.assertRaisesRegex(
            ce, 'time course', i.model,
            os.path.join(DIR, 'ch-13-no-time-course.channelml'))

        # Unreadable steady-state expression
        with WarningCollector() as w:
            m = i.model(
                os.path.join(DIR, 'ch-14-inf-bad-expression.channelml'))
        self.assertRaises(myokit.MissingRhsError, m.validate)
        self.assertEqual(w.count(), 1)
        self.assertIn('expression for steady state', w.text())
        self.assertEqual('mill', m.get('Nav1_3.h.inf').meta['expression'])

        # Unreadable time-constant expression
        with WarningCollector() as w:
            m = i.model(
                os.path.join(DIR, 'ch-15-tau-bad-expression.channelml'))
        self.assertRaises(myokit.MissingRhsError, m.validate)
        self.assertEqual(w.count(), 1)
        self.assertIn('expression for time course', w.text())
        self.assertEqual('house', m.get('Nav1_3.h.tau').meta['expression'])

        # No gates
        self.assertRaisesRegex(
            ce, 'at least one gate', i.model,
            os.path.join(DIR, 'ch-16-no-gates.channelml'))

        # Invalid name
        with WarningCollector() as w:
            m = i.model(
                os.path.join(DIR, 'ch-17-invalid-name.channelml'))
        m.validate()
        self.assertEqual(w.count(), 1)
        self.assertIn('Invalid name', w.text())

    def test_c_style_if(self):
        # Test parsing of a c-style cond?then:else expression.

        # If-then-else
        i = myokit.formats.importer('channelml')
        m = i.model(
            os.path.join(DIR, 'ch-18-c-style-if.channelml'))
        self.assertEqual(
            str(m.get('Nav1_3.m.alpha').rhs()),
            'if(membrane.v != -26, 12, 13)'
        )

        # If-then
        self.assertEqual(
            str(m.get('Nav1_3.m.beta').rhs()),
            'if(membrane.v != -26, 100, 0)'
        )


if __name__ == '__main__':
    unittest.main()



#
# Flattens HTML and attempts to create readable ASCII code.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import html
import html.parser
import re
import textwrap


def html2ascii(html, width=79, indent='  '):
    """
    Flattens HTML and attempts to create readable ASCII code.

    The output will be text-wrapped after ``width`` characters. Each new level
    of nesting will be indented with the text given as ``indent``.
    """
    # Create asciifier and go!
    f = Asciifier(width, indent)
    f.feed(html)
    f.close()
    return f.get_text()


class Asciifier(html.parser.HTMLParser):
    INDENT = 1
    DEDENT = -1
    WHITE = [' ', '\t', '\f', '\r', '\n']

    def __init__(self, line_width=79, indent='  '):
        super().__init__(convert_charrefs=False)

        # In <head> yes/no
        self.inhead = False

        self.text = []  # Current document
        self.line = []  # Current (unwrapped) line

        # Ordered/unorderd lists
        self.limode = []
        self.licount = []

        # Output options
        self.LW = line_width
        self.TAB = indent

    def endline(self):
        """
        End the current line.
        """
        line = ''.join(self.line)
        self.line = []
        if line:
            self.text.append(line)

    def blankline(self):
        """
        Inserts a blank line.
        """
        i = -1
        last = self.text[-1:]
        while last in [[self.INDENT], [self.DEDENT]]:
            i -= 1
            last = self.text[i:1 + i]
        if last != ['']:
            self.text.append('')

    def handle_data(self, data):
        """
        Handle text between tags
        """
        # Ignore content while in head
        if self.inhead:
            return

        # Convert to ascii and back to remove all non-ascii content
        data = data.encode('ascii', 'ignore').decode('ascii')

        # Non-empty? Then add to line
        if data:
            self.line.append(data)

    def handle_starttag(self, tag, attrs):
        """ Called when a start tag is encountered. """
        if ':' in tag:
            tag = tag[1 + tag.index(':'):]

        if tag == 'head':
            self.inhead = True

        elif tag == 'p':
            self.endline()
            self.blankline()

        elif tag == 'h1':
            self.endline()
            self.blankline()
            self.text.append('=' * self.LW)

        elif tag == 'h2':
            self.endline()
            self.blankline()
            self.text.append('-' * self.LW)

        elif tag == 'h3':
            self.endline()
            self.blankline()
            self.text.append('.' * self.LW)

        elif tag == 'ul' or tag == 'ol':
            self.endline()
            self.blankline()
            self.text.append(self.INDENT)
            self.limode.append(tag)
            self.licount.append(0)

        elif tag == 'li':
            self.endline()
            # Get mode and count, or use default if not set
            if len(self.limode):
                limode = self.limode[-1]
                licount = self.licount[-1]
                self.licount[-1] += 1
            else:
                limode = 'ul'
                licount = 0

            # Add bullet
            if limode == 'ul':
                self.line.append('* ')
            else:
                self.line.append(str(1 + licount) + u' ')

        elif tag == 'em' or tag == 'i':
            self.line.append(' *')

        elif tag == 'strong' or tag == 'b':
            self.line.append(' **')

        elif tag == 'u':
            self.line.append(' _')

    def handle_startendtag(self, tag, attrs):
        """ Called when an open/end tag is encountered. """
        if ':' in tag:
            tag = tag[1 + tag.index(':'):]

        if tag == 'br':
            self.endline()

        elif tag == 'hr':
            self.endline()
            self.text.append('-' * self.LW)
            self.endline()

    def handle_endtag(self, tag):
        """ Called when an endtag is encountered. """
        if ':' in tag:
            tag = tag[1 + tag.index(':'):]

        if tag == 'head':
            self.inhead = False

        elif tag == 'p':
            self.endline()
            self.blankline()

        elif tag == 'h1':
            self.endline()
            self.text.append('=' * self.LW)
            self.blankline()

        elif tag == 'h2':
            self.endline()
            self.text.append('-' * self.LW)
            self.blankline()

        elif tag == 'h3':
            self.endline()
            self.text.append('.' * self.LW)
            self.blankline()

        elif tag == 'ul' or tag == 'ol':
            self.endline()
            self.text.append(self.DEDENT)
            self.blankline()

            if len(self.limode):
                self.limode.pop()
            if len(self.licount):
                self.licount.pop()

        elif tag == 'li':
            pass

        elif tag == 'em' or tag == 'i':
            self.line.append('* ')

        elif tag == 'strong' or tag == 'b':
            self.line.append('** ')

        elif tag == 'u':
            self.line.append('_ ')

    def handle_entityref(self, name):
        # Ignore content while in head
        if self.inhead:
            return

        # Convert html characters
        data = html.unescape('&' + name + ';')

        # Convert to ascii and back to strip out non-ascii content
        data = data.encode('ascii', 'ignore').decode('ascii')

        # Non-empty? Then add to line
        if data:
            self.line.append(data)

    def get_text(self):
        self.endline()
        buf = []
        pre = ''
        n = self.LW
        ntab = len(self.TAB)
        dent = 0
        white = '[' + ''.join(self.WHITE) + ']+'
        space = ' '
        next_can_be_blank = False
        for line in self.text:
            if line == self.INDENT:
                dent += 1
                pre = dent * self.TAB
                n -= ntab
            elif line == self.DEDENT:
                dent -= 1
                pre = dent * self.TAB
                n += ntab
            else:
                line = re.sub(white, space, line).strip()
                if line == '':
                    if next_can_be_blank:
                        buf.append('')
                        next_can_be_blank = False
                else:
                    buf.extend([pre + x for x in textwrap.wrap(line, n)])
                    next_can_be_blank = True
        return ('\n'.join(buf)).strip()




#!/usr/bin/env python3
#
# Tests the CellML 2.0 writer.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import re
import unittest

import myokit
import myokit.formats.cellml.v2 as cellml

from myokit.tests import TemporaryDirectory, WarningCollector


class TestCellMLWriter(unittest.TestCase):
    """ Tests for cellml.Writer. """

    def test_component(self):
        # Tests if components are written

        m1 = cellml.Model('m')
        c = m1.add_component('c')
        d = m1.add_component('d')

        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        self.assertIn('c', m2)
        self.assertIn('d', m2)

    def test_connections(self):
        # Tests if connections are written

        m1 = cellml.Model('m')
        c = m1.add_component('c')
        d = m1.add_component('d')
        p = c.add_variable('p', 'mole', 'public')
        q = d.add_variable('q', 'mole', 'public')
        r = c.add_variable('r', 'ampere', 'public')
        s = d.add_variable('r', 'ampere', 'public')
        p.set_initial_value(1)
        s.set_initial_value(2)
        m1.add_connection(p, q)
        m1.add_connection(r, s)

        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        p = m2['c']['p']
        q = m2['d']['q']
        self.assertIn(p, q.connected_variables())
        self.assertIn(q, p.connected_variables())
        self.assertIs(p.initial_value_variable(), p)
        self.assertIs(q.initial_value_variable(), p)

        r = m2['c']['r']
        s = m2['d']['r']
        self.assertIn(r, s.connected_variables())
        self.assertIn(s, r.connected_variables())
        self.assertIs(r.initial_value_variable(), s)
        self.assertIs(s.initial_value_variable(), s)

    def test_evaluating_derivatives(self):
        # Writes and then parses a model and compares the derivatives

        m1 = myokit.load_model('example')
        c1 = cellml.Model.from_myokit_model(m1)
        xml = cellml.write_string(c1)
        c2 = cellml.parse_string(xml)
        m2 = c2.myokit_model()

        self.assertEqual(
            m1.get('membrane.V').eval(), m2.get('membrane.V').eval())
        self.assertEqual(m1.get('ina.m').eval(), m2.get('ina.m').eval())
        self.assertEqual(m1.get('ina.h').eval(), m2.get('ina.h').eval())
        self.assertEqual(m1.get('ina.j').eval(), m2.get('ina.j').eval())
        self.assertEqual(m1.get('ica.d').eval(), m2.get('ica.d').eval())
        self.assertEqual(m1.get('ica.f').eval(), m2.get('ica.f').eval())
        self.assertEqual(m1.get('ik.x').eval(), m2.get('ik.x').eval())
        self.assertEqual(m1.get('ica.Ca_i').eval(), m2.get('ica.Ca_i').eval())

    def test_encapsulation(self):
        # Tests if encapsulation relationships are written out.

        # Create a model with
        #      d
        #   a     e
        #       f   c
        #             b
        #
        m1 = cellml.Model('m')
        a = m1.add_component('a')
        b = m1.add_component('b')
        c = m1.add_component('c')
        d = m1.add_component('d')
        e = m1.add_component('e')
        f = m1.add_component('f')
        a.set_parent(d)
        e.set_parent(d)
        f.set_parent(e)
        c.set_parent(e)
        b.set_parent(c)

        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)
        a, b, c, d, e, f = [m2.component(x) for x in 'abcdef']
        self.assertIs(a.parent(), d)
        self.assertIs(e.parent(), d)
        self.assertIs(f.parent(), e)
        self.assertIs(c.parent(), e)
        self.assertIs(b.parent(), c)

    def test_initial_value_representation(self):
        # Test the way initial values are represented in generated CellML code

        def find(xml):
            i = xml.index(b'initial_value')
            i += 15
            j = xml.index(b'"', i)
            return xml[i:j].decode()

        m = cellml.Model('m', '2.0')
        c = m.add_component('c')
        p = c.add_variable('p', 'mole')

        p.set_initial_value(1.234)
        x = find(cellml.write_string(m))
        self.assertEqual(x, '1.234')

        p.set_initial_value(1e-6)
        x = find(cellml.write_string(m))
        self.assertEqual(x, '1e-06')

        p.set_initial_value(1e9)
        x = find(cellml.write_string(m))
        self.assertEqual(x, '1.00000000000000000e+09')

        # String e+00
        p.set_initial_value(1.23424352342423)
        x = find(cellml.write_string(m))
        self.assertEqual(x, '1.23424352342422994')

    def test_maths(self):
        # Test maths is written

        self._test_maths('2.0')

    def _test_maths(self, version):
        # Test maths is written (in selected ``version``)

        # Create model
        m1 = cellml.Model('m', version)
        c1 = m1.add_component('c')
        p1 = c1.add_variable('p', 'mole')
        p1.set_initial_value(2)
        q1 = c1.add_variable('q', 'dimensionless')
        r1 = c1.add_variable('r', 'second')
        r1.set_initial_value(0.1)
        t1 = c1.add_variable('t', 'second')
        m1.set_variable_of_integration(t1)

        # Add component without maths
        d1 = m1.add_component('d')
        s1 = d1.add_variable('s', 'volt')
        s1.set_initial_value(1.23)

        # Add two equations
        # Note: Numbers without units become dimensionless in CellML
        eq1 = myokit.Equation(
            myokit.Name(q1),
            myokit.Plus(myokit.Number(3, myokit.units.mole), myokit.Name(p1)))
        er1 = myokit.Equation(
            myokit.Derivative(myokit.Name(r1)),
            myokit.Power(myokit.Name(q1), myokit.Number(2)))
        q1.set_equation(eq1)
        r1.set_equation(er1)

        # Write and read
        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        # Check results
        p2, q2, r2, s2 = m2['c']['p'], m2['c']['q'], m2['c']['r'], m2['d']['s']
        subst = {
            myokit.Name(p1): myokit.Name(p2),
            myokit.Name(q1): myokit.Name(q2),
            myokit.Name(r1): myokit.Name(r2),
        }
        eq2 = eq1.clone(subst)
        er2 = er1.clone(subst)

        self.assertEqual(q2.equation(), eq2)
        self.assertEqual(r2.equation(), er2)
        self.assertEqual(
            s2.initial_value(), myokit.Number(1.23, myokit.units.volt))
        self.assertFalse(p2.is_state())
        self.assertFalse(q2.is_state())
        self.assertTrue(r2.is_state())
        self.assertFalse(s2.is_state())
        self.assertIs(m2.variable_of_integration(), m2['c']['t'])

    def test_model(self):
        # Tests model writing

        m1 = cellml.Model('model_name')
        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        self.assertEqual(m2.name(), 'model_name')
        self.assertEqual(len(m2), 0)

    def test_ordering(self):
        # Tests ordering of components and variables

        # Check component ordering
        m = cellml.Model('m')
        c = m.add_component('C')
        a = m.add_component('A')
        d = m.add_component('D')
        b = m.add_component('B')

        xml = cellml.write_string(m)
        reg = re.compile(b'<component [^>]*name="[\\w]+"')
        items = reg.findall(xml)
        items_sorted = list(sorted(items))
        self.assertEqual(items, items_sorted)

        # Check variable ordering
        m = cellml.Model('m')
        c = m.add_component('C')
        r = c.add_variable('r', 'volt')
        p = c.add_variable('p', 'ampere')
        s = c.add_variable('s', 'kilogram')
        q = c.add_variable('q', 'mole')
        r.set_initial_value(1)
        p.set_initial_value(2)
        q.set_initial_value(3)
        s.set_initial_value(4)

        xml = cellml.write_string(m)
        reg = re.compile(b'<variable [^>]*name="[\\w]+"')
        items = reg.findall(xml)
        items_sorted = list(sorted(items))
        self.assertEqual(items, items_sorted)

    def test_units(self):
        # Test writing of units

        u1 = myokit.parse_unit('kg*m^2/mole^3 (0.123)')

        m1 = cellml.Model('mmm')
        m1.add_units('flibbit', u1)
        m1.add_units('special_volt', myokit.units.Volt)
        d = m1.add_component('ddd')
        q = d.add_variable('q', 'flibbit')
        q.set_equation(myokit.Equation(
            myokit.Name(q), myokit.Number(2, u1)))

        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        q = m2['ddd']['q']
        self.assertEqual(q.units().name(), 'flibbit')
        self.assertEqual(q.units().myokit_unit(), u1)
        self.assertEqual(
            m2.find_units('special_volt').myokit_unit(), myokit.units.volt)

        # Dimensionless units with a multiplier
        u1 = myokit.parse_unit('1 (0.123)')
        m1 = cellml.Model('mmm')
        m1.add_units('flibbit', u1)
        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)
        u2 = m2.find_units('flibbit')
        u2 = u2.myokit_unit()
        self.assertEqual(u1, u2)

    def test_variable(self):
        # Tests writing of variables

        m1 = cellml.Model('m')
        c = m1.add_component('c')
        p = c.add_variable('p', 'mole')
        q = c.add_variable('q', 'kelvin', interface='public')
        r = c.add_variable('r', 'ampere', interface='private')
        p.set_initial_value(1)

        with WarningCollector():
            xml = cellml.write_string(m1)
            m2 = cellml.parse_string(xml)

        p, q, r = m2['c']['p'], m2['c']['q'], m2['c']['r']
        self.assertEqual(p.units().name(), 'mole')
        self.assertEqual(q.units().name(), 'kelvin')
        self.assertEqual(r.units().name(), 'ampere')
        self.assertEqual(p.interface(), 'none')
        self.assertEqual(q.interface(), 'public')
        self.assertEqual(r.interface(), 'private')
        self.assertEqual(
            p.initial_value(), myokit.Number(1, myokit.units.mole))
        self.assertEqual(q.initial_value(), None)
        self.assertEqual(r.initial_value(), None)

    def test_version(self):
        # Tests if all supported versions can be written

        # Version 2.0
        m1 = cellml.Model('m', version='2.0')
        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)
        self.assertEqual(m2.version(), '2.0')

    def test_write_file(self):
        # Tests write_file

        m1 = cellml.Model('ernie')
        with TemporaryDirectory() as d:
            path = d.path('test.cellml')
            cellml.write_file(path, m1)
            m2 = cellml.parse_file(path)

        self.assertEqual(m2.name(), 'ernie')
        self.assertEqual(len(m2), 0)

    def test_write_string(self):
        # Tests write_string

        m1 = cellml.Model('ernie')
        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        self.assertEqual(m2.name(), 'ernie')
        self.assertEqual(len(m2), 0)


if __name__ == '__main__':
    import warnings
    warnings.simplefilter('always')
    unittest.main()



#
# Export as an EasyML model.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import collections
import os
import warnings

import myokit.formats
import myokit.lib.guess as guess
import myokit.lib.hh as hh
import myokit.lib.markov as markov


class EasyMLExporter(myokit.formats.Exporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` generates a ``.model``
    file in the ``EasyML`` format used by CARP/CARPEntry.

    For details of the language, see
    https://carpentry.medunigraz.at/carputils/examples/tutorials/tutorials.01_EP_single_cell.04_limpet_fe.run.html
    """ # noqa
    # To test the output, you'll need `$ limpet_fe.py model_file`

    def model(self, path, model, protocol=None, convert_units=True):
        """
        Exports a :class:`myokit.Model` in EasyML format, writing the result to
        the file indicated by ``path``.

        A :class:`myokit.ExportError` will be raised if any errors occur.

        Arguments:

        ``path``
            The path to write the generated model to.
        ``model``
            The :class:`myokit.Model` to export.
        ``protocol``
            A protocol - this will be ignored!
        ``convert_units``
            If set to ``True``, the metohd will attempt to convert to CARP's
            preferred units for voltage (mV), current (A/F), and time (ms).

        """
        import myokit.formats.easyml as easyml

        # Test model is valid
        try:
            model.validate()
        except myokit.MyokitError:
            raise myokit.ExportError('EasyML export requires a valid model.')

        # Rewrite model so that any Markov models have a 1-sum(...) state
        # This also clones the model, so that changes can be made
        model = markov.convert_markov_models_to_compact_form(model)

        # Replace any RHS references to state derivatives with references to
        # intermediary variables
        model.remove_derivative_references()

        # Remove hardcoded stimulus protocol, if any
        guess.remove_embedded_protocol(model)

        # List of variables not to output
        ignore = set()

        # Find membrane potential
        vm = guess.membrane_potential(model)

        # Ensure vm is a state, so that expressions depending on V are not seen
        # as constants.
        if not vm.is_state():
            vm.promote(-80)

        # Don't output vm
        ignore.add(vm)

        # Get time variable
        time = model.time()

        # Find currents (must be done before i_ion is removed)
        currents = guess.membrane_currents(model)

        # Unit conversion
        def convert(var, unit, helpers=None):
            if var.unit() is None:
                return
            try:
                var.convert_unit(unit, helpers=helpers)
            except myokit.IncompatibleUnitError:
                warnings.warn(
                    'Unable to convert ' + var.qname() + ' to recommended'
                    ' units of ' + str(unit) + '.')

        # Use recommended units
        time_factor = time_factor_inv = None
        if convert_units:
            recommended_current_units = [
                myokit.parse_unit('pA/pF'),
                myokit.parse_unit('uA/cm^2'),
            ]
            helpers = []
            membrane_capacitance = guess.membrane_capacitance(model)
            if membrane_capacitance is not None:
                helpers.append(membrane_capacitance.rhs())
            convert(vm, myokit.units.mV)
            for current in currents:
                if current.unit() not in recommended_current_units:
                    convert(current, 'A/F', helpers=helpers)

            # Get conversion factor for time
            # Multiplying by time_factor = Going from time_units to ms
            time_unit = time.unit()
            if time_unit is not None and time_unit != myokit.units.ms:
                try:
                    time_factor = myokit.Number(myokit.Unit.conversion_factor(
                        time_unit, myokit.units.ms))
                    time_factor_inv = myokit.Number(
                        1 / time_factor.eval(),
                        1 / time_factor.unit())
                except myokit.IncompatibleUnitError:
                    warnings.warn(
                        'Unable to convert time units ' + str(time_unit)
                        + ' to recommended units of ms.')

        # Remove time, pacing variable, diffusion current, and i_ion
        pace = model.binding('pace')
        i_diff = model.binding('diffusion_current')
        i_ion = model.label('cellular_current')
        for var in [time, pace, i_diff, i_ion]:
            if var is None:
                continue

            # Replace references to these variables with 0
            refs = list(var.refs_by())
            subst = {var.lhs(): myokit.Number(0)}
            for ref in refs:
                ref.set_rhs(ref.rhs().clone(subst=subst))

            # Remove variable
            var.parent().remove_variable(var)

            # Remove from currents, if present
            try:
                currents.remove(var)
            except ValueError:
                pass

        # Remove unused variables

        # Start by setting V's rhs to the sum of currents, so all currents
        # count as used
        rhs = myokit.Number(0)
        for v in currents:
            rhs = myokit.Plus(rhs, v.lhs())
        vm.set_rhs(rhs)

        # Remove all bindings and labels (so they register as unused)
        for b, var in model.bindings():
            var.set_binding(None)
        for b, var in model.labels():
            var.set_label(None)

        # And add the time variable back in
        time = vm.parent().add_variable(time.name())
        time.set_rhs(0)
        time.set_binding('time')
        ignore.add(time)

        # Remove unused
        model.validate(remove_unused_variables=True)

        # Find special variables
        hh_states = set()
        alphas = {}
        betas = {}
        taus = {}
        infs = {}

        # If an inf/tau/alpha/beta is used by more than one state, we create a
        # copy for each user, so that we can give the variables the appropriate
        # name (e.g. x_inf, y_inf, z_inf) etc.
        def renamable(var):
            srefs = [v for v in var.refs_by() if v.is_state()]
            if len(srefs) > 1:
                v = var.parent().add_variable_allow_renaming(var.name())
                v.set_rhs(myokit.Name(var))
                v.set_unit(var.unit())
                return v
            return var

        # Find HH state variables and their infs, taus, alphas, betas
        for var in model.states():
            ret = hh.get_inf_and_tau(var, vm)
            if ret is not None:
                ignore.add(var)
                hh_states.add(var)
                infs[renamable(ret[0])] = var
                taus[renamable(ret[1])] = var
                continue
            ret = hh.get_alpha_and_beta(var, vm)
            if ret is not None:
                ignore.add(var)
                hh_states.add(var)
                alphas[renamable(ret[0])] = var
                betas[renamable(ret[1])] = var
                continue

        hh_variables = (set(alphas.keys()) | set(betas.keys()) |
                        set(taus.keys()) | set(infs.keys()))

        # Find Markov models (must be run before time conversion)
        markov_models = markov.find_markov_models(model)

        # Add time conversion factor for states and HH variables
        if time_factor is not None:
            for var in model.states():
                var.set_rhs(myokit.Multiply(var.rhs(), time_factor_inv))
            for var in alphas:
                var.set_rhs(myokit.Multiply(var.rhs(), time_factor_inv))
            for var in betas:
                var.set_rhs(myokit.Multiply(var.rhs(), time_factor_inv))
            for var in taus:
                var.set_rhs(myokit.Multiply(var.rhs(), time_factor))

        # Create variable names
        # The following strategy is followed:
        #  - HH variables are named after their state
        #  - All other variables are checked for special names, and changed if
        #    necessary
        #  - Any conflicts are resolved in a final step

        # Detect names with special meanings
        def special_start(name):
            if name.startswith('diff_') or name.startswith('d_'):
                return True
            if name.startswith('alpha_') or name.startswith('a_'):
                return True
            if name.startswith('beta_') or name.startswith('b_'):
                return True
            if name.startswith('tau_'):
                return True
            return False

        def special_end(name):
            return name.endswith('_init') or name.endswith('_inf')

        # Create initial variable names
        var_to_name = collections.OrderedDict()
        for var in model.variables(deep=True, sort=True):

            # Delay naming of HH variables until their state has a name
            if var in hh_variables:
                continue

            # Start from simple variable name
            name = var.name()

            # Add parent if needed
            if name in ['alpha', 'beta', 'inf', 'tau']:
                name = var.parent().name() + '_' + name

            # Avoid names with special meaning
            if special_end(name):
                name += '_var'
            if special_start(name):
                name = var.parent().name() + '_' + name
                if special_start(name):
                    name = var.qname().replace('.', '_')
                if special_start(name):
                    name = 'var_' + name

            # Store (initial) name for var
            var_to_name[var] = name

        # Check for conflicts with known keywords
        from . import keywords
        reserved = [
            'Iion',
        ]
        needs_renaming = collections.OrderedDict()
        for keyword in keywords:
            needs_renaming[keyword] = []
        for keyword in reserved:
            needs_renaming[keyword] = []

        # Find naming conflicts, create inverse mapping
        name_to_var = collections.OrderedDict()
        for var, name in var_to_name.items():

            # Known conflict?
            if name in needs_renaming:
                needs_renaming[name].append(var)
                continue

            # Test for new conflicts
            var2 = name_to_var.get(name, None)
            if var2 is not None:
                needs_renaming[name] = [var2, var]
            else:
                name_to_var[name] = var

        # Resolve naming conflicts
        for name, variables in needs_renaming.items():
            # Add a number to the end of the name, increasing it until it's
            # unique
            i = 1
            root = name + '_'
            for var in variables:
                name = root + str(i)
                while name in name_to_var:
                    i += 1
                    name = root + str(i)
                var_to_name[var] = name
                name_to_var[name] = var

        # Remove reverse mappings
        del name_to_var, needs_renaming

        # Create names for HH variables
        for var, state in alphas.items():
            var_to_name[var] = 'alpha_' + var_to_name[state]
        for var, state in betas.items():
            var_to_name[var] = 'beta_' + var_to_name[state]
        for var, state in taus.items():
            var_to_name[var] = 'tau_' + var_to_name[state]
        for var, state in infs.items():
            var_to_name[var] = var_to_name[state] + '_inf'

        # Create naming function
        def lhs(e):
            if isinstance(e, myokit.Derivative):
                return 'diff_' + var_to_name[e.var()]
            elif isinstance(e, myokit.LhsExpression):
                return var_to_name[e.var()]
            elif isinstance(e, myokit.Variable):
                return var_to_name[e]
            raise ValueError(   # pragma: no cover
                'Not a variable or LhsExpression: ' + str(e))

        # Create expression writer
        e = easyml.EasyMLExpressionWriter()
        e.set_lhs_function(lhs)

        # Test if can write in dir (raises exception if can't)
        self._test_writable_dir(os.path.dirname(path))

        # Write equations
        eol = '\n'
        eos = ';\n'
        with open(path, 'w') as f:
            # Write meta data
            f.write('/*' + eol)
            f.write('This file was generated by Myokit.' + eol)
            if model.meta:
                f.write(eol)
                for key, val in sorted(model.meta.items()):
                    f.write(key + ': ' + val + eol)
                f.write(eol)
            f.write('*/' + eol + eol)

            # Write membrane potential
            f.write(lhs(vm) + '; .nodal(); .external(Vm);' + eol)

            # Write current
            f.write('Iion; .nodal(); .external();' + eol + eol)

            # Write initial conditions
            for v in model.states():
                f.write(lhs(v) + '_init = '
                        + myokit.float.str(v.initial_value(True)) + eos)
            f.write(eol)

            # Write remaining variables
            for c in model.components(sort=True):
                todo = c.variables(deep=True, sort=True)
                todo = [v for v in todo if v not in ignore]
                if todo:
                    f.write('// ' + c.name() + eol)
                    for v in todo:
                        f.write(e.eq(v.eq()) + eos)
                    f.write(eol)

            # Write sum of currents variable
            f.write('// Sum of currents' + eol)
            f.write('Iion = ')
            f.write(' + '.join([lhs(v) for v in currents]))
            f.write(eos + eol)

            # Write solution methods for Markov models
            markov_states = set()
            for states in markov_models:
                f.write('// Markov model' + eol)
                f.write('group {' + eol)
                for state in states:
                    if state.is_state():
                        f.write('  ' + lhs(state) + eos)
                        markov_states.add(state)
                f.write('}.method(markov_be)' + eos + eol)

            # Solve all non-Markovian and non-HH variables with CVODE
            cvode_states = [
                x for x in model.states() if not (
                    x == vm or x in markov_states or x in hh_states)]
            if cvode_states:
                f.write(
                    '// Solve non-HH and non-Markov states with CVODE' + eol)
                f.write('group {' + eol)
                for v in cvode_states:
                    f.write('  ' + lhs(v) + eos)
                f.write('}.method(cvode)' + eos + eol)

            # Trace all currents and state variables
            f.write('// Trace all currents and state variables' + eol)
            f.write('group {' + eol)
            for v in currents:
                f.write('  ' + lhs(v) + eos)
            for v in model.states():
                f.write('  ' + lhs(v) + eos)
            f.write('}.trace()' + eos + eol)

            # Make all constants parameters
            parameters = list(model.variables(const=True, sort=True))
            if parameters:
                f.write('// Parameters' + eol)
                f.write('group {' + eol)
                for p in parameters:
                    f.write('  ' + lhs(p) + eos)
                f.write('}.param()' + eos + eol)

    def supports_model(self):
        """ See :meth:`myokit.formats.Exporter.supports_model()`. """
        return True




#
# Loads settings from configuration file in user home directory or attempts to
# guess best settings.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import configparser
import os
import platform
import sys
import warnings

# Load Myokit, at least, the bit that's been setup so far. This just means
# this method will add a link to the myokit module already being loaded
# into this method's namespace. This allows us to use the constants defined
# before this method was called.
import myokit


def _create(path):
    """
    Attempts to guess the best settings and stores them in a new configuration
    file.
    """
    # Get operating system
    system = platform.system()

    # Create config parser
    config = configparser.RawConfigParser(allow_no_value=True)

    # Make the parser case sensitive (need for unix paths!)
    config.optionxform = str

    # General information
    config.add_section('myokit')
    config.set(
        'myokit',
        '# This file can be used to set global configuration options for'
        ' Myokit.')

    # Compatibility settings
    config.add_section('compatibility')
    config.set(
        'compatibility',
        '# Optional settings to make Myokit work on tricky systems.')
    config.set('compatibility', '# Don\'t capture compiler output.')
    config.set('compatibility', '#no_capture = True')
    config.set('compatibility', '# Don\'t use the file-descriptor method.')
    config.set('compatibility', '#no_fd_capture = True')

    # Date format
    config.add_section('time')
    config.set('time', '# Date format used throughout Myokit')
    config.set('time', '# The format should be acceptable for time.strftime')
    config.set('time', 'date_format', myokit.DATE_FORMAT)
    config.set('time', '# Time format used throughout Myokit')
    config.set('time', '# The format should be acceptable for time.strftime')
    config.set('time', 'time_format', myokit.TIME_FORMAT)

    # GUI Backend
    config.add_section('gui')
    config.set('gui', '# Backend to use for graphical user interface.')
    config.set('gui', '# Valid options are pyqt6, pyqt5, pyside6 and pyside2.')
    config.set('gui', '# Leave unset for automatic selection.')
    config.set('gui', '#backend = pyqt6')
    config.set('gui', '#backend = pyqt5')
    config.set('gui', '#backend = pyside6')
    config.set('gui', '#backend = pyside2')

    # Locations of sundials library
    config.add_section('sundials')
    config.set(
        'sundials', '# Location of sundials shared libary files'
        ' (.so, .dll, or .dylib).')
    config.set('sundials', '# Multiple paths can be set using ; as separator.')

    if system == 'Windows':     # pragma: no linux cover
        # Windows: Don't set (see load())
        config.set('sundials', '#lib', ';'.join([
            'C:\\Program Files\\sundials\\lib',
            'C:\\Program Files (x86)\\sundials\\lib',
        ]))
    elif system == 'Darwin':    # pragma: no linux cover
        # Apple
        # Standard: /usr/local/lib
        # Macports: /opt/local/lib
        # Homebrew: /opt/homebrew
        config.set('sundials', 'lib', ';'.join([
            '/usr/local/lib',
            '/usr/local/lib64',
            '/opt/local/lib',
            '/opt/local/lib64',
            '/opt/homebrew/lib',
            '/opt/homebrew/lib64',
        ]))
    else:
        # Linux
        config.set('sundials', 'lib', ';'.join([
            '/usr/local/lib',
            '/usr/local/lib64',
        ]))

    config.set('sundials',
               '# Location of sundials header files (.h).')
    config.set('sundials', '# Multiple paths can be set using ; as separator.')

    if system == 'Windows':     # pragma: no linux cover
        # Windows: Don't set (see load())
        config.set('sundials', '#inc', ';'.join([
            'C:\\Program Files\\sundials\\include',
            'C:\\Program Files (x86)\\sundials\\include',
        ]))
    elif system == 'Darwin':    # pragma: no linux cover
        # Apple
        # Standard: /usr/local/include
        # Macports: /opt/local/include
        # Homebrew: /opt/homebrew/include
        config.set('sundials', 'inc', ';'.join([
            '/usr/local/include',
            '/opt/local/include',
            '/opt/homebrew/include',
        ]))
    else:
        # Linux
        config.set('sundials', 'inc', ';'.join([
            '/usr/local/include',
        ]))

    # Locations of OpenCL libraries
    config.add_section('opencl')
    config.set('opencl',
               '# Location of opencl shared libary files (.so, .dll, .dylib).')
    config.set('opencl', '# Multiple paths can be set using ; as separator.')

    if system == 'Windows':     # pragma: no linux cover
        # All windowses
        c64 = 'C:\\Program Files\\'
        config.set('opencl', 'lib', ';'.join([
            c64 + 'Intel\\OpenCL SDK\\6.3\\lib\\x64',
            c64 + 'AMD APP SDK\\2.9\\bin\\x64',
            c64 + 'NVIDIA GPU Computing Toolkit\CUDA\\v11.8\\lib\\x64',
        ]))
    else:
        # Linux and mac
        paths = [
            '/usr/lib64',
            '/usr/lib64/nvidia',
            '/usr/local/cuda/lib64',
        ]
        if system == 'Darwin':  # pragma: no linux cover
            paths.append('/System/Library/Frameworks')
        config.set('opencl', 'lib', ';'.join(paths))

    config.set('opencl', '# Location of opencl header files (.h).')
    config.set('opencl', '# Multiple paths can be set using ; as separator.')

    if system == 'Windows':     # pragma: no linux cover
        # All windowses
        c64 = 'C:\\Program Files\\'
        config.set('opencl', 'inc', ';'.join([
            c64 + 'Intel\\OpenCL SDK\\6.3\\include',
            c64 + 'AMD APP SDK\\2.9\\include',
            c64 + 'NVIDIA GPU Computing Toolkit\\CUDA\\v11.8\\include',
        ]))
    else:
        # Linux and mac
        paths = [
            '/usr/include/CL',
            '/usr/local/cuda/include',
        ]
        if system == 'Darwin':  # pragma: no linux cover
            paths.append('/System/Library/Frameworks')
        config.set('opencl', 'inc', ';'.join(paths))

    # Write ini file
    try:
        with open(path, 'w') as configfile:
            config.write(configfile)
    except IOError:     # pragma: no cover
        warnings.warn('Warning: Unable to write settings to ' + str(path))


def _load():
    """
    Reads the configuration file and attempts to set the library paths.
    """
    # Location of configuration file
    path = os.path.join(myokit.DIR_USER, 'myokit.ini')

    # No file present? Create one and return
    if not os.path.isfile(path):
        _create(path)

    # In Python <3.2, strings like "x ; y" are treated as "x" followed by a
    # comment. These shouldn't appear in myokit ini files!
    if sys.hexversion < 0x03020000:     # pragma: no cover
        with open(path, 'r') as f:
            lines = f.readlines()

        import re
        inline_comment = re.compile('[\w]+[\s]*=[\s]*.+?\s+(;)')
        for i, line in enumerate(lines):
            m = inline_comment.match(line)
            if m is not None:
                x = m.start(1) - 1
                raise ImportError(
                    'Unsupported syntax found in ' + str(path) + ' on line '
                    + str(1 + i) + ', character ' + str(x) + ', semicolons (;)'
                    + ' must not be preceded by whitespace: ```'
                    + line.strip() + '```.')
        del lines, inline_comment

    # Create the config parser (no value allows comments)
    config = configparser.RawConfigParser(allow_no_value=True)

    # Make the parser case sensitive (need for unix paths!)
    config.optionxform = str

    # Parse the config file
    config.read(path)

    # Compatibility options
    if config.has_option('compatibility', 'no_capture'):
        x = config.get('compatibility', 'no_capture').strip().lower()
        if x == 'true':
            myokit.COMPAT_NO_CAPTURE = True
        elif x == 'false':
            myokit.COMPAT_NO_CAPTURE = False
        elif x != '':
            warnings.warn(
                'Invalid setting in myokit.ini. Expected values for no_capture'
                ' are true, false, or not set (empty), but got: ' + x)

    if config.has_option('compatibility', 'no_fd_capture'):
        x = config.get('compatibility', 'no_fd_capture').strip().lower()
        if x == 'true':
            myokit.COMPAT_NO_FD_CAPTURE = True
        elif x == 'false':
            myokit.COMPAT_NO_FD_CAPTURE = False
        elif x != '':
            warnings.warn(
                'Invalid setting in myokit.ini. Expected values for'
                ' no_fd_capture are true, false, or not set (empty), but got: '
                + x)

    # Date format
    if config.has_option('time', 'date_format'):
        x = config.get('time', 'date_format')
        if x:
            myokit.DATE_FORMAT = str(x)

    # Time format
    if config.has_option('time', 'time_format'):
        x = config.get('time', 'time_format')
        if x:
            myokit.TIME_FORMAT = str(x)

    # GUI Backend
    if config.has_option('gui', 'backend'):
        x = config.get('gui', 'backend').strip().lower()
        if x == 'pyqt6':
            myokit.FORCE_PYQT6 = True
            myokit.FORCE_PYQT5 = False
            myokit.FORCE_PYSIDE6 = False
            myokit.FORCE_PYSIDE2 = False
        elif x == 'pyqt5':
            myokit.FORCE_PYQT6 = False
            myokit.FORCE_PYQT5 = True
            myokit.FORCE_PYSIDE6 = False
            myokit.FORCE_PYSIDE2 = False
        elif x == 'pyside6':
            myokit.FORCE_PYQT6 = False
            myokit.FORCE_PYQT5 = False
            myokit.FORCE_PYSIDE6 = True
            myokit.FORCE_PYSIDE2 = False
        elif x == 'pyside2':
            myokit.FORCE_PYQT6 = False
            myokit.FORCE_PYQT5 = False
            myokit.FORCE_PYSIDE6 = False
            myokit.FORCE_PYSIDE2 = True
        elif x != '':
            warnings.warn(
                'Invalid setting in myokit.ini. Expected values for backend'
                ' are pyqt6, pyqt5, pyside6, or pyside2. Got: ' + x)

    # Sundials libraries, header files, and version
    if config.has_option('sundials', 'lib'):
        myokit.SUNDIALS_LIB.extend(_path_list(config.get('sundials', 'lib')))
    if config.has_option('sundials', 'inc'):
        myokit.SUNDIALS_INC.extend(_path_list(config.get('sundials', 'inc')))

    # Dynamically add embedded sundials paths for windows
    if platform.system() == 'Windows':  # pragma: no linux cover
        _dynamically_add_embedded_sundials_win()

    # OpenCL libraries and header files
    if config.has_option('opencl', 'lib'):
        myokit.OPENCL_LIB.extend(_path_list(config.get('opencl', 'lib')))
    if config.has_option('opencl', 'inc'):
        myokit.OPENCL_INC.extend(_path_list(config.get('opencl', 'inc')))


def _dynamically_add_embedded_sundials_win():   # pragma: no linux cover
    """
    On windows, sundials binaries are packaged with Myokit. Storing this
    location in the config file could cause issues, for example if a user moves
    a Myokit installation, reinstalls e.g. Anaconda to a different path, uses
    multiple Myokits or Python distros etc. So instead of storing this location
    in the config file, it is added dynamically.
    This is only done _if_ the user doesn't set an explicit path (i.e. we allow
    user overrides).
    """
    sundials_win = os.path.abspath(
        os.path.join(myokit.DIR_DATA, 'sundials-win-vs'))
    if len(myokit.SUNDIALS_LIB) == 0:
        myokit.SUNDIALS_LIB.append(os.path.join(sundials_win, 'lib'))
    if len(myokit.SUNDIALS_INC) == 0:
        myokit.SUNDIALS_INC.append(os.path.join(sundials_win, 'include'))


def _path_list(text):
    return [
        os.path.expandvars(os.path.expanduser(x))
        for x in [x.strip() for x in text.split(';')] if x != '']


# Load settings
_load()



#!/usr/bin/env python3
#
# Tests the system() method
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit


class SystemInfoTest(unittest.TestCase):
    """
    Tests the system info method.
    """

    def test_system_info(self):
        import matplotlib
        matplotlib.use('template')

        self.assertIsInstance(myokit.system(), str)
        with myokit.tools.capture():
            myokit.system(live_printing=True)


if __name__ == '__main__':
    unittest.main()
